"""
A collection of tools for working with finance data in an academic setting. Compare and summarize data sets,
propensity score matching, analyze outliers and missing data, and create/reformat several types of data sets,
including: difference-in-difference, Fama-French-style factors, and portfolios.
"""
import pandas as pd
import os, datetime, warnings, sys

from pd_utils import expand_time, cumulate, convert_sas_date_to_pandas_date, factor_reg_by, load_sas, \
                    long_to_wide, year_month_from_date, join_col_strings
    
# from ..compustat import convert_gvkey, load_compustat, merge_compustat
# from ..config import data_path


def replace_missing_csv(csv_list, missing_rep):
    '''
    Replaces missing items in a CSV with a given missing representation string.
    '''
    full_list = []
    for line in csv_list:
        line_list = line.split(',')
        new_line_list = []
        for item in line_list:
            if item == '': #if the item is missing
                item = missing_rep
            new_line_list.append(item)
        full_list.append(new_line_list)
    return full_list

def merge_dsenames(df, on='TICKER', get='PERMNO', date='Date', 
                   other_byvars=None, 
                   crsp_dir=None):
    '''
    Merges with dsenames file on on variable (TICKER, PERMNO, PERMCO, NCUSIP, CUSIP6), to get get variable (same list).
    Must have a Date variable in df.
    
    Default is to match on TICKER and pull PERMNO.
    
    Required inputs:
    df: pandas dataframe containing any of (TICKER, PERMNO, PERMCO, NCUSIP, CUSIP6)
    
    Optional inputs:
    on: str, column name to merge on, one of (TICKER, PERMNO, PERMCO, NCUSIP, CUSIP6)
    get: str or list, column or columns to get from dsenames, any of (TICKER, PERMNO, PERMCO, NCUSIP, CUSIP6)
         that aren't already in on
    date: str, column name of date variable
    other_byvars: any other variables signifying groups in the data, prevents from collapsing those groups
    '''

    # Set default CRSP dir
    # if crsp_dir is None:
    #     crsp_dir = data_path('CRSP')

    #Make get a list
    if isinstance(get, str):
        get = [get]
    assert isinstance(get, list)
    
    #Make other byvars a list
    if not other_byvars:
        other_byvars = []
    if isinstance(other_byvars, str):
        other_byvars = [other_byvars]
    assert isinstance(other_byvars, list)
    
    assert on not in get #can't get what we already have
    
    #Pull from CRSP dsenames file
    file = 'dsenames'
    fullpath = os.path.join(crsp_dir, file + '.sas7bdat')
    names_df = load_sas(fullpath)
    
    #Convert NCUSIP to CUSIP6
    if on == 'CUSIP6' or 'CUSIP6' in get:
        names_df['CUSIP6'] = names_df['NCUSIP'].apply(lambda x: x if pd.isnull(x) else x[:6])
    
    names_df['start'] = convert_sas_date_to_pandas_date(names_df['NAMEDT'])
    names_df['end'] = convert_sas_date_to_pandas_date(names_df['NAMEENDT'])
    names_df['end'] = names_df['end'].fillna(datetime.date.today())
    
    
    #Now perform merge
    merged = df[[on, date] + other_byvars].merge(names_df[['start','end', on] + get], how='left', on=on)
    #Drop out observations not in name date range
    valid = (merged[date] >= merged['start']) & (merged[date] <= merged['end'])
    #However if there is not a match, doing merged[valid] would drop the observation instead of leaving nan
    #Therefore, take merged[valid] and merge back again to original
    new_merged = df.merge(merged[valid].drop(['start','end'],axis=1), how='left', on=[on, date] + other_byvars)
    new_merged = new_merged.reset_index(drop=True)
    
    if 'PERMNO' in get:
        #Dsenames has no record of which permno is the primary link when a firm has multiple share classes.
        #To get this information, we must merge ccmxpf_linktable. We want to keep only primary links.
        dups = new_merged[[date, on] + other_byvars].duplicated(keep=False) #series of True or False of whether duplicated row
        if dups.any(): #this means we got more than one permno for a single period/firm/byvars
            duplicated = new_merged[dups].reset_index() #puts index in a column for later use
            not_duplicated = new_merged[~dups]
            
            #Take duplicated, merge to ccmxpf_linktable to get gvkey, and the ones which do not have gvkeys are
            #the non-primary links
            with_gvkey = get_gvkey_or_permno(duplicated, date) #default is to get gvkey with permno
            removed_duplicates = with_gvkey[~pd.isnull(with_gvkey['GVKEY'])].drop('GVKEY', axis=1)
            
            #Set index back
            removed_duplicates.set_index('index', inplace=True)
            
            #Now append back together and sort
            full = not_duplicated.append(removed_duplicates)
            new_merged = full.sort_index()
    
    return new_merged.reset_index(drop=True)

def get_gvkey_or_permno(df, datevar, get='GVKEY', other_byvars=None,
                        crsp_dir=None):
    """
    Takes a dataframe containing either GVKEY or PERMNO and merges to the CRSP linktable to get the other one.
    """

    # Set default CRSP dir
    # if crsp_dir is None:
    #     crsp_dir = data_path('CRSP')

    if get == 'GVKEY':
        rename_get = 'gvkey'
        l_on = 'PERMNO'
        r_on = 'lpermno'
    elif get == 'PERMNO':
        rename_get = 'lpermno'
        l_on = 'GVKEY'
        r_on = 'gvkey'
    else:
        raise ValueError('Need get="GVKEY" or "PERMNO"')
        
    #Make other byvars a list
    if not other_byvars:
        other_byvars = []
    if isinstance(other_byvars, str):
        other_byvars = [other_byvars]
    assert isinstance(other_byvars, list)
    
    link_name = 'ccmxpf_linktable.sas7bdat'
    link_path = os.path.join(crsp_dir, link_name)
    
    link = load_sas(link_path)
    link['linkdt'] = convert_sas_date_to_pandas_date(link['linkdt'])
    link['linkenddt'] = convert_sas_date_to_pandas_date(link['linkenddt'])
    #If end date is missing, that means link is still active. Make end date today.
    link['linkenddt'] = link['linkenddt'].fillna(datetime.date.today())
    
    #Remove links with no permno so that they don't match to nans in the input df
    link.dropna(subset=['lpermno'], inplace=True)
    
    merged = df.merge(link[['lpermno','gvkey', 'linkdt', 'linkenddt','linkprim']], how='left', left_on=l_on, right_on=r_on)

    valid = (merged[datevar] >= merged.linkdt) & \
            (merged[datevar] <= merged.linkenddt) & \
            (merged.linkprim == 'P')
        
    merged = merged[valid].drop(['linkdt','linkenddt', 'linkprim', r_on], axis=1).drop_duplicates()
    merged.rename(columns={rename_get:get}, inplace=True)
    
    #Now merge back to the original again to ensure that rows are not deleted
    new_merged = df.merge(merged[['PERMNO','GVKEY', datevar] + other_byvars],
                          how='left', on=[l_on, datevar] + other_byvars)
    
    return new_merged


class GetCRSP:
    
    def __init__(self, debug=False, crsp_dir=None):

        # Set default CRSP dir
        # if crsp_dir is None:
        #     crsp_dir = data_path('CRSP')

        self.debug = debug
        self.crsp_dir = crsp_dir
        self.loaded_m = False
        self.loaded_d = False
        self.crsp_dfs = {} #container for both monthly and daily crsp df
        
    
    def pull_crsp(self, df, coid='PERMNO', freq='m', get=['PRC','SHROUT'], date='Date', other_byvars=None,
                 time=None, between_time=False, wide=True,
                 abret=False, window=None, cumret=False, includefac=False, includecoef=False, 
                 drop_first=False):
        '''
        Pulls prices and returns from CRSP. Currently supports the monthly file merging on PERMNO.

        WARNING: Will overwrite variables called "Year" and "Month" if merging the monthly file

        coid = string, company identifier (currently supports 'GVKEY', 'TICKER', 'PERMNO', 'PERMCO', 'NCUSIP')
        freq = 'm' for monthly, 'd' for daily
        get = 'PRC', 'RET', 'SHROUT', 'VOL', 'CFACPR', 'CFACSHR', or a list combining any of these
        date = name of datetime column in quotes
        other_byvars = other by vars in dataset besides company identifier and date. Used for long_to_wide
        time = list of integers or None. If not None, will pull the variables for a time difference of the time
               numbers given, and put them in wide format in the output dataset. For instance time=[-12,0,12] with
               freq='m' and get='RET' would pull three returns per observation, one twelve months prior, one contemporaneous,
               and one twelve months later, naming them RET-12, RET0, and RET12. If freq='d' in the same example, it would be
               twelve days prior, etc.
        between_time = bool. Only applicable if time is not None. True to generate periods in between given time periods, e.g.
                        passing time=[12,24,36] will get periods 12, 13, 14, ..., 35, 36.
        abret = 1, 3, 4, or False. If 1, 3, or 4 is passed and 'RET' is in get, will calculate abnormal returns according to CAPM,
                3 or 4 factor model, respectively. 
        window = Integer or None. Must provide an integer if abret is not False. This is the number of prior periods to use
                 for estimating the loadings in factor models. For instance, if freq='m', window=36 would use the prior
                 three years of returns to estimate the loadings.
        cumret = 'between', 'zero', 'first', or False. time must not be None and RET in get for this option to matter. 
                 When pulling returns for multiple periods, gives the option to cumulate returns. If False, will just return 
                 returns for the individual periods. 
                 If 'zero', will give returns since the original date. 
                 If 'between', will give returns since the prior requested time period.
                 If 'first', will give returns since the first requested time period.
                 For example, if our input data was for date 1/5/2006, and in the CRSP table we had:
                     permno  date       RET
                     10516   1/5/2006   10%
                     10516   1/6/2006   20%
                     10516   1/7/2006    5%
                     10516   1/8/2006   30%
                 Then get_crsp(df, time=[1,3], get='RET', cumret=None) would return:
                     permno  date       RET1  RET3
                     10516   1/5/2006   20%   30%
                 Then get_crsp(df, time=[1,3], get='RET', cumret='between') would return:
                     permno  date       RET1  RET3
                     10516   1/5/2006   20%   36.5%
                 Then get_crsp(df, time=[1,3], get='RET', cumret='zero') would return:
                     permno  date       RET1  RET3
                     10516   1/5/2006   20%   63.8%
                 The output for cumret='first' would be the same as for cumret='zero' because the first period is period zero.
                 Had time been =[-1, 1, 3], then returns would be calculated from period -1 to period 1, and period -1 to period 3. 
        includefac = Boolean, True to include factors in output
        includecoef = Boolean, True to include factor coefficients in output
        wide = True for output data to be wide form, False for long form. Only applies when time is not None.
        drop_first = bool, set to True to drop observations for first time. Can only be used when time != None, and
                     when cumret != False. This is a
                     convenience function for estimating cumulative return windows. For example, if time = [-1, 1], 
                     then the typical output would include both cum_RET-1 and cum_RET1. All we actually care about is the
                     cumulative return over the window, which is equal to cum_RET1. drop_first=True will drop out 
                     RET-1 and cum_RET-1 from the output.
        debug: bool, set to True to restrict CRSP to only PERMNOs 10516 (gvkey=001722) and 10517 (gvkey=001076)

        Typical usage:
        Calculating a return window with abnormal returns included:
            get_crsp(df, get='RET', freq='d', time=[-1, 1], cumret='between', abret=4, window=250, drop_first=True)
        '''  
        self.df = df.copy()
        self.coid = coid
        self.freq = freq
        self.get = get
        self.date = date
        self.other_byvars = other_byvars
        self.time = time
        self.intermediate_periods = between_time
        self.wide = wide
        self.abret = abret
        self.window = window
        self.cumret = cumret
        self.includefac = includefac
        self.includecoef = includecoef
        self.drop_first = drop_first
        
        self._log('Initializing pull_crsp function')
    
        self._clear_old_variables()
        self._check_inputs()
        self._load_crsp()
        
        if self.time == None and self.abret == False: 
            return self._merge_crsp(self.df, self.date) #if we're not doing anything special, just merge CRSP
        
        if self.time:
            self._handle_time()
        else:
            self._handle_no_time()
            
        self._create_key()
        
        if self.abret:
            self._handle_abret()
        else:
            self.long_df = self._merge_crsp(self.long_df, date='Shift Date') #time but not abret        
        
        if self.cumret:
            self._handle_cumret()
        
        self.long_df.drop('key', axis=1, inplace=True)
        
        if self.wide:
            return self._long_to_wide()
        else:
            return self._output_long()
    
    def _clear_old_variables(self):
        """
        If pull_crsp is run multiple times, values of variables would still be set
        """
        try: del self.long_df
        except AttributeError: #long_df not set yet
            pass
        
        
    
    def _output_long(self):
        if self.debug:
            return self.long_df
        for var in ['Shift','Shift Date']:
            try:
                self.long_df.drop(var, axis=1, inplace=True)
            except ValueError:
                pass
        return self.long_df
    
    def _long_to_wide(self):
        self._log('Reshaping long to wide.')
        byvars = [self.coid, self.date]
        if self.other_byvars: #add other byvars for correct long_to_wide
            byvars += self.other_byvars
#         if self.freq == 'm':
#             self.long_df.drop(['Year','Month'], axis=1, inplace=True)
        try:
            self.long_df.drop('Shift Date', axis=1, inplace=True)
        except ValueError: #this is the case where time was not shifted
            #Therefore need to create a colindex which is just 0 for the current time period
            self.long_df['Shift'] = 0
        
        widedf = long_to_wide(self.long_df.reset_index(drop=True),
                              groupvars=byvars, values=self.get, colindex='Shift')
        #chop off zeros from contemporaneous
        return widedf.rename(columns={name: name[:-1] for name in widedf.columns \
                                      if name.endswith('0') and name[:name.find('0')] in self.get}) 
    
    def _handle_cumret(self):
        self._log('Cumret detected.')
        cumvars = ['RET']
        if self.abret: cumvars += ['ABRET']
        self.get += ['cum_' + str(c) for c in cumvars] #get will be used in the end for pivot, need to add pivoting variables
        with warnings.catch_warnings(): #cumulate will raise a warning if time is supplied when method is not between
            warnings.simplefilter('ignore') #suppress that warning
            self._log('Cumulating returns with method {} for time {}.'.format(self.cumret, self.time))
            byvars = ['PERMNO', self.date]
            if self.other_byvars:
                byvars += self.other_byvars
                
            self.long_df = cumulate(self.long_df, cumvars, periodvar='Shift Date', method=self.cumret,
                               byvars=byvars, time=self.time, grossify=True)
            
            
            
            ###########TEMP
#             import pdb
#             pdb.set_trace()
            
            
            
            ############
            
        #Now need to remove unneeded periods
        # First check if we should be getting intermediate periods, e.g. [1, 4] -> [1, 2, 3, 4]
        if self.intermediate_periods:
            keep_time = [t for t in range(min(self.time), max(self.time) + 1)]
        else:
            keep_time = self.time

        if self.drop_first:
            keep_time = keep_time[1:]
        self.long_df = self.long_df[self.long_df['Shift'].isin(keep_time)]
            
    def _handle_abret(self):
        self._log('Abret detected.')
        assert isinstance(self.abret, int) and not isinstance(self.abret, bool) #True evaluates as int
        assert isinstance(self.window, int)
        
        facs = ['mktrf']
        if self.abret >= 3:
            facs += ['hml','smb']
        if self.abret == 5:
            facs += ['rmw', 'cma']
        if self.abret not in (1,3,5):
            raise ValueError('Currently only supports 1, 3, and 5-factor models (mktrf, hml, smb, rmw, cma)')
        
        prior_wind = [i for i in range(0,-self.window - 1,-1)] #e.g. if window = 3, prior_wind = [0, -1, -2, -3]
        self._log('Creating abret window periods.')
        self.long_df = expand_time(self.long_df, datevar=self.newdate, 
                                   freq=self.freq, time=prior_wind, 
                                   newdate='Window Date', shiftvar='Wind')
        self._log('Merging with CRSP to get abret window data as well as regular data.')       
        self.long_df = self._merge_crsp(df=self.long_df, date='Window Date') 
        self._log('Getting Fama-French factors, then running regressions.')
        self.long_df = get_abret(self.long_df, 'key', fulldatevar='Window Date', freq=self.freq,
                                 abret_fac=self.abret, includecoef=self.includecoef,
                                 includefac=self.includefac)
        
        self._log('Dropping unneeded observations (abret window dates).')
        if self.time:
            #keep only observations for shift dates
            self.long_df = self.long_df[self.long_df['Window Date'] == self.long_df['Shift Date']].reset_index(drop=True) 
        else: #didn't get multiple times per period
            #keep only original observations
            self.long_df = self.long_df[self.long_df['Window Date'] == self.long_df[self.date]].reset_index(drop=True)
        self.get += ['ABRET'] #get will be used in the end for pivot, need to add pivoting variables
        
        drop = ['Wind', 'Window Date'] #variables to be dropped
        coefs = ['coef_' + fac for fac in facs]
        if self.includefac: self.get += facs + ['rf']
        if self.includecoef: self.get += coefs + ['const']
        self.long_df.drop(drop, axis=1, inplace=True)
        
    
    def _create_key(self):
        if self.other_byvars:
            self.byvars += self.other_byvars
        self.long_df['key'] = join_col_strings(self.long_df,self.byvars)
    
    def _handle_time(self):
        self._log('Time detected.')
        #Next section only runs if time is not None
        #Ensure time is of the right type
        if isinstance(self.time, int): self.time = [self.time]
        assert (isinstance(self.time, list) and isinstance(self.time[0], int))
        if self.cumret:
            self._log('Cumret detected, will generate intermediate periods.')
            intermediate_periods = True
        else:
            intermediate_periods = self.intermediate_periods
        self._log('Generating periods {} {}'.format(self.time, '+ itermediate' if intermediate_periods else ''))
        self.long_df = expand_time(self.df, intermediate_periods=intermediate_periods,
                                   datevar=self.date, freq=self.freq, time=self.time)
        self._log('Finished generating periods. Generating key.')
        self.byvars = ['PERMNO', self.date, 'Shift Date']
        self.newdate = 'Shift Date'
        
    def _handle_no_time(self):
        self.long_df = self.df.copy()
        self._log('Generating key.')
        self.byvars = ['PERMNO', self.date]
        self.newdate = self.date
    
    def _check_inputs(self):
        self._log('Checking inputs.')
        
#         if self.debug:
#             self._log('All inputs: ')
#             self._log(str(self.__dict__))

        #Check get
        if isinstance(self.get, list):
            self.get = [item.upper() for item in self.get]
#             for item in self.get:
#                 assert item in ['PRC','RET','SHROUT','VOL','CFACPR','CFACSHR']
        elif isinstance(self.get, str):
            self.get = self.get.upper()
#             assert self.get in ['PRC','RET','SHROUT','VOL','CFACPR','CFACSHR']
            self.get = [self.get]
        else:
            raise ValueError('''Get should be a list or str containing 'PRC','RET','SHROUT','VOL','CFACPR', or 'CFACSHR'.''')

        #Check to make sure inputs make sense
        assert not ((self.abret == False) and (self.window != None)) #can't specify window without abret
        assert not ((self.abret != False) and (self.window == None)) #must specify window with abret
        assert not ((self.abret == False) and (self.includefac == True)) #cannot include factors unless calculating abnormal returns
        assert not ((self.abret == False) and (self.includecoef == True)) #cannot include factor coefs unless calculating abnormal returns
        assert not ((self.abret != False) and ('RET' not in self.get)) #can't calculate abnormal returns without getting returns
        assert not (self.cumret and ('RET' not in self.get)) #can't cumulate returns without getting returns
        assert not (self.cumret and (self.time == None)) #can't cumulate over a single period
        assert not ((self.drop_first == True) and (self.time == None)) #can't drop first shifted time period if there are none
        assert not ((self.drop_first == True) and (len(self.time) == 1)) #can't drop first shifted time period if there's only one
        assert not ((self.drop_first == True) and (self.cumret == False)) #no reason to drop first shifted time period if we're not cumulating


        #Check to make sure company identifier is valid
        assert self.coid in ('GVKEY','TICKER', 'PERMNO', 'PERMCO', 'NCUSIP', 'CUSIP6')
        if self.coid != 'PERMNO':
            self._log('Company ID is not PERMNO. Getting PERMNO.')
            if self.coid == 'GVKEY':
                self.df = get_gvkey_or_permno(self.df, self.date, get='PERMNO',
                                              other_byvars=self.other_byvars) #grabs permno from ccmxpf_linktable
                self._log('Pulled PERMNO from ccmxpf_linktable.')
            else: #all others go through dsenames
                self.df = merge_dsenames(self.df, on=self.coid, date=self.date,
                                        other_byvars=self.other_byvars) #grabs permno
                self._log('Pulled PERMNO from dsenames.')

        #Check to make sure no columns currently in the dataframe have the same name as columns
        #we will be adding
        for col in self.get:
            if col in self.df.columns:
                self.df.rename(columns={col: col + '_old'}, inplace=True)

        #Ensure other_byvars is a list
        if isinstance(self.other_byvars, str):
            self.other_byvars = [self.other_byvars]
        assert isinstance(self.other_byvars, (list, type(None)))
    
    def _load_crsp(self):
        if self.freq.lower() == 'm':
            if not self.loaded_m:
                self.__load_crsp()
            self.loaded_m = True
        if self.freq.lower() == 'd':
            if not self.loaded_d:
                self.__load_crsp()
            self.loaded_d = True
        
    def __load_crsp(self):
        #Check frequency
        if self.freq.lower() == 'm':
            filename = 'msf'
        elif self.freq.lower() == 'd':
            filename = 'dsf'
        else: raise ValueError('use m or d for frequency')
        if self.debug: filename += '_test' #debug datasets only have permnos 10516, 10517

        #Load in CRSP file
        self._log('Loading CRSP dataframe...')
        filepath = os.path.join(self.crsp_dir, filename + '.sas7bdat')
        self.crsp_dfs[self.freq.lower()] = load_sas(filepath)
        self._log('Loaded.')

        #Change date to datetime format
        self._log('Converting SAS date to Pandas format.')
        self.crsp_dfs[self.freq.lower()]['DATE'] = convert_sas_date_to_pandas_date(
                                                    self.crsp_dfs[self.freq.lower()]['DATE'])
        self._log('Converted.')
    
    def _log(self, message):
        if message != '\n':
            time = datetime.datetime.now().replace(microsecond=0)
            message = str(time) + ': ' + message
        sys.stdout.write(message + '\n')
        sys.stdout.flush()
    
    def _merge_crsp(self, df, date):
        self._log('Merging CRSP to dataframe.')
        df = df.copy()

        #If we are using the monthly file, we need to merge on month and year
        if self.freq.lower() == 'm':
            self.crsp_df = self.crsp_dfs['m']
            self.crsp_df = year_month_from_date(self.crsp_df, date='DATE')
            orig_df_for_merge = year_month_from_date(df, date=date)

            #Now perform merge
            merged = orig_df_for_merge.merge(self.crsp_df[['Month','Year','PERMNO'] + self.get],
                                   how='left', on=['PERMNO','Month','Year'])
            merged.drop(['Month','Year'], axis=1, inplace=True)
            
        if self.freq.lower() == 'd':
            self.crsp_df = self.crsp_dfs['d']
            merged = df.merge(self.crsp_df[['DATE','PERMNO'] + self.get],
                              how='left', right_on=['PERMNO','DATE'],
                              left_on=['PERMNO', date])
            merged.drop('DATE', axis=1, inplace=True)
            
        self._log('Completed merge.')

        #Temp
        return merged

    
def get_ff_factors(df, fulldatevar=None, year_month=None, freq='m',
                   subset=None, ff_df=None, ff_dir=None,
                  custom_ff_name=None):
    """
    Pulls Fama-French factors and merges them to dataset
    
    df: Input dataframe
    fulldatevar: String name of date variable to merge on. Specify this OR year and month variable. Must use this
                 and not year_month if pulling daily factors. If merging with monthly factors, will create month
                 and year variables in the output dataset. Warning: Will overwrite any variables called Month
                 and Year in the input data.
    year_month: Two element list of ['yearvar','monthvar']. Specify this OR full date variable.
    freq: 'm' for monthly factors, 'd' for daily
    subset: str or list, names of ff factors to pull. Can specify any of 'mktrf', 'smb', 'hml', 'rmw', 'cma', 'umd'
    ff_df: pandas DataFrame containing factors. if passed, will ignore ff_dir and custom_ff_name
    ff_dir: folder containing FF data
    custom_ff_name: used in combination with ff_dir to select file
    """

    # Set default FF dir
    # if ff_dir is None:
    #     ff_dir = data_path('FF')
   
    #Make sure inputs are correct
    freq = _set_freq(freq)
    assert isinstance(df, pd.DataFrame)
    assert isinstance(ff_dir, str)
    assert not (fulldatevar == None and year_month == None)
    assert not (fulldatevar == None and freq == 'd')
    
    df = df.copy()
    
    if not subset:
        subset = ['mktrf', 'smb', 'hml', 'rmw', 'cma']
    if isinstance(subset, str):
        subset = [subset]
    assert isinstance(subset, list)
    for item in subset:
        assert item in ['mktrf', 'smb', 'hml', 'rmw', 'cma', 'umd']
        
    subset = subset.copy() #don't modify original beyond converting to list
    
    if freq == 'm': 
        ff_name = 'ff_fac_month.csv'
        drop = False
        if year_month != None:
            left_datevars = year_month
            df_for_merge = df
        else: #fulldatevar specified
            df_for_merge = year_month_from_date(df, date=fulldatevar)
            left_datevars = ['Year','Month']
            drop = True
        right_datevars = ['year','month']
    else: 
        drop = False
        df_for_merge = df
        ff_name = 'ff_fac_daily.csv'
        left_datevars = fulldatevar
        right_datevars = ['date']
        
    subset += right_datevars + ['rf'] #need to pull date variables and risk free rate as well
    
    if custom_ff_name is not None:
        ff_name = custom_ff_name

    # Use passed df. If not passed, load df.
    if ff_df is None:
        path = os.path.join(ff_dir, ff_name)
        ff_df = _load_data_by_extension_and_convert_date(path, freq=freq)
    
    merged = df_for_merge.merge(ff_df[subset], how='left', left_on=left_datevars, right_on=right_datevars)
    merged.drop(right_datevars, axis=1, inplace=True)
    
    if drop:
        merged.drop(left_datevars, axis=1, inplace=True)
    
    return merged

def _load_data_by_extension_and_convert_date(filepath, freq='m'):
    filename, file_extension = os.path.splitext(filepath)
    extension = file_extension.lower()
    if extension == '.sas7bdat':
        df = load_sas(filepath)
        if freq != 'm':
            df['date'] = convert_sas_date_to_pandas_date(df['date']) #convert to date object
        return df
    elif extension == '.csv':
        if freq != 'm':
            return pd.read_csv(filepath, parse_dates=['date'])
        else:
            return pd.read_csv(filepath)
    else:
        raise ValueError(f'Please pass a sas7bdat or csv for FF factors, got {extension}')

def _set_freq(freq: str) -> str:
    compare_freq = freq.lower()
    if compare_freq in ('m', 'month', 'months', 'monthly'):
        return 'm'
    else:
        # All other frequencies handled the same way
        return freq

def get_abret(df, byvars, fulldatevar='Date', year_month=None, freq='m', abret_fac=5, retvar='RET',
              includecoef=False, includefac=False, mp=False, stderr=False, **get_ff_kwargs):
    """
    Takes a dataframe containing a column of returns, dates, and at least one by variable and calculates
    abnormal returns for the model of choice. Returns a dataframe with the abnormal returns merged.
    
    Required inputs:
    df: pandas dataframe
    byvars: str or list of strs, column names of columns on which to form by groups

    Optional inputs:
    fulldatevar: str, name of column containing date variable. If provided, don't provide year_month.
    year_month: list of strs, columns names of year and month variables, e.g. ['Year','Month']. Must
                set fulldatevar to None if year_month is provided.
    freq: 'm' or 'd', 'm' for monthly returns, 'd' for daily returns
    abret_fac: int (1, 3, 5), abnormal return model
    retvar: str, name of return variable
    includecoef: bool, set to True to get factor loadings
    includefac: bool, set to True to get factors and risk free rate
    mp: False to use single processor, True to use all processors, int to use # processors
    stderr: bool, True to include standard errors of coefficients

    """
    assert abret_fac in (1, 3, 5)
    factors = ['mktrf']
    if abret_fac >= 3:
        factors += ['smb','hml']
    if abret_fac == 5:
        factors += ['rmw','cma']
    
    out = get_ff_factors(
        df, 
        fulldatevar=fulldatevar,
        freq=freq,
        subset=factors,
        year_month=year_month,
        **get_ff_kwargs
    )
    out = factor_reg_by(out, byvars, fac=abret_fac, retvar=retvar, mp=mp, stderr=stderr)

    if not includefac:
        out.drop(factors + ['rf'], axis=1, inplace=True)
    if not includecoef:
        out.drop(['const'] + ['coef_' + fac for fac in factors], axis=1, inplace=True)
        
    return out

# def load_and_merge_compustat(df, get=['sale'], freq='a', gvkeyvar='gvkey', datevar='Date', debug=False,
#                              comp_dir=None):
#     """
#     Convenience function for both loading and merging compustat to existing dataframe.
#
#     Required inputs:
#     df: Pandas dataframe containing a date variable and gvkey
#
#     Optional inputs:
#     get: List of strings, variable names to pull from compustat, not including those needed for
#          the merge. Do not add the 'q' or 'y' for quarterly variables, this is done automatically.
#          'y' variables will be converted to quarterly values by looking at changes.
#     freq: string, 'a' or 'q' for annual or quarterly compustat (funda, fundq)
#     gvkeyvar: string, name of variable containing gvkeys
#     datevar: string, name of date variable in df on which to merge. Will pull the newest data
#              that is before or on this date.
#     debug: bool, True to restrict to only gvkeys (001076, 001722)
#     comp_dir: string, directory containing compustat files
#     """
#
    # Set default FF dir
    # if comp_dir is None:
    #     comp_dir = data_path('Compustat')

    # convert_gvkey(df, gvkeyvar)
    # comp = load_compustat(freq, get=get, debug=debug, comp_dir=comp_dir)
    # return merge_compustat(df, comp, datevar=datevar).rename(columns={'gvkey':gvkeyvar})


from datacode.psm.typing import StrOrNone

accurate_prediction_name = 'Accurate Prediction'
diff_var = 'Diff'
t_stat_var = '$t$-stat'
matched_var = 'Matched'

accuracy_pct = '% Accuracy'

def get_prob_treated_varname(treated_var: str, prob_treated_var_name: StrOrNone = None) -> str:
    if prob_treated_var_name is None:
        return prob_treated_varname(treated_var)

    return prob_treated_var_name

def prob_treated_varname(treated_var: str) -> str:
    return f'Prob({treated_var})'

def predict_treated_varname(treated_var: str) -> str:
    return f'Predicted {treated_var}'


import pandas as pd
import pd_utils

def _long_counts_to_wide_df(long_df: pd.DataFrame, col: str, row_byvar: str, col_byvar: str) -> pd.DataFrame:
    return pd_utils.long_to_wide(
        long_df[[row_byvar, col_byvar, col]],
        groupvars=[row_byvar],
        values=[col],
        colindex=[col_byvar],
        colindex_only=True
    ).set_index(row_byvar).fillna(0)


from typing import List, Sequence, Optional, Union

import pandas as pd
import pyexlatex as pl

from datacode.sem.constants import STANDARD_SCALE_MESSAGE, ROBUST_SCALE_MESSAGE


def summary_latex_table(fit_df: pd.DataFrame, structural_dfs: Sequence[pd.DataFrame],
                        latent_dfs: Sequence[pd.DataFrame], scale: bool, robust_scale: bool,
                        equations: Optional[Sequence[pl.Equation]] = None,
                        begin_below_text: Optional[str] = None,
                        end_below_text: Optional[str] = None, replace_below_text: Optional[str] = None,
                        caption: str = 'Structural Equation Model (SEM)') -> pl.Table:
    below_text = _below_text(
        scale,
        robust_scale,
        equations=equations,
        begin_below_text=begin_below_text,
        end_below_text=end_below_text,
        replace_below_text=replace_below_text
    )
    panels = _summary_panels(fit_df, structural_dfs, latent_dfs)
    table = pl.Table.from_panel_list(
        panels,
        caption=caption,
        below_text=below_text
    )
    return table


def _below_text(scale: bool, robust_scale: bool, equations: Optional[Sequence[pl.Equation]] = None,
                begin_below_text: Optional[str] = None,
                end_below_text: Optional[str] = None, replace_below_text: Optional[str] = None) -> Union[str, List[str]]:
    if replace_below_text:
        return replace_below_text

    below_text = []
    if begin_below_text:
        below_text.append(begin_below_text)
    if equations:
        below_text.append('The model is represented by the following equations:')
        below_text.extend(equations)
    if scale:
        below_text.append(STANDARD_SCALE_MESSAGE)
    if robust_scale:
        below_text.append(ROBUST_SCALE_MESSAGE)
    if end_below_text:
        below_text.append(end_below_text)

    return below_text


def _summary_panels(fit_df: pd.DataFrame, structural_dfs: Sequence[pd.DataFrame],
                    latent_dfs: Sequence[pd.DataFrame]) -> List[pl.Panel]:
    fit_panel = pl.Panel.from_df_list([fit_df], name='Goodness of Fit', include_index=True, include_columns=False)
    structural_dts = _coefficient_data_tables(structural_dfs)
    structural_panel = pl.Panel.from_data_tables(structural_dts, name='Structural Equations', )
    latent_dts = _coefficient_data_tables(latent_dfs)
    latent_panel = pl.Panel.from_data_tables(latent_dts, name='Latent Equations')
    return [fit_panel, structural_panel, latent_panel]


def _coefficient_data_tables(coef_dfs: Sequence[pd.DataFrame]) -> List[pl.DataTable]:
    dts = []
    for coef_df in coef_dfs:
        fmt_coef_df = coef_df.applymap(lambda x: f'{x:.3f}' if not pd.isnull(x) else x).fillna('')
        dt = pl.DataTable.from_df(fmt_coef_df, include_index=True, extra_header=coef_df.index.name)
        dts.append(dt)
    return dts



import pandas as pd
from semopy import Optimizer, inspect

from datacode.utils.suppress import no_stdout


def lambda_df_from_opt(opt: Optimizer) -> pd.DataFrame:
    with no_stdout():
        matrix_results = inspect(opt, mode='mx')
    lambda_df = matrix_results[1][1]
    return lambda_df



import itertools


def partial(func, *args, **keywords):
    """
    A version of functools.partial that allows specifying positional arguments out of order using ellipses.

    See: https://stackoverflow.com/a/11173903

    Examples:
        >>> print partial(pow, ..., 2, 3)(5) # (5^2)%3
        1
        >>> print partial(pow, 2, ..., 3)(5) # (2^5)%3
        2
        >>> print partial(pow, 2, 3, ...)(5) # (2^3)%5
        3
        >>> print partial(pow, 2, 3)(5) # (2^3)%5
        3

    """
    def newfunc(*fargs, **fkeywords):
        newkeywords = keywords.copy()
        newkeywords.update(fkeywords)
        return func(*(newfunc.leftmost_args + fargs + newfunc.rightmost_args), **newkeywords)
    newfunc.func = func
    args = iter(args)
    newfunc.leftmost_args = tuple(itertools.takewhile(lambda v: v != Ellipsis, args))
    newfunc.rightmost_args = tuple(args)
    newfunc.keywords = keywords
    return newfunc



import re
from typing import Tuple

from datacode.portfolio.typing import PortfolioPair, PortfolioName


def _differenced_port_name(difference_port: PortfolioPair =('High', 'Low')) -> str:
    return f'{difference_port[0]} - {difference_port[1]}'


def _is_differenced_port_name(portname: PortfolioName) -> bool:
    pattern = re.compile(r'[\d\w]+ \- [\d\w]+')
    if pattern.fullmatch(str(portname)):
        return True
    else:
        return False


def _ports_from_differenced_port_name(portname: PortfolioName) -> Tuple[str, str]:
    long_port, short_port = str(portname).split(' - ')
    return long_port, short_port


import datetime
from typing import Callable, Optional, Any, Dict

from datacode.models.pipeline.operations.operation import DataOperation, OperationOptions
from datacode.models.source import DataSource


class GenerationOperation(DataOperation):
    """
    Data operation that takes one DataSource as an input and outputs a DataSource
    """
    num_required_sources = 0
    options: 'GenerationOptions'
    result: 'DataSource'

    def __init__(self, options: 'GenerationOptions', **result_kwargs):
        super().__init__(
            [],
            options,
            **result_kwargs
        )

    def _execute(self):
        ds = self.options.func(**self.options.func_kwargs)
        self.result.update_from_source(ds)
        return self.result

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        # TODO [#46]: better summary for DataGeneratorPipeline
        print(f'Calling function {self.options.func.__name__} with kwargs {self.options.func_kwargs} to '
              f'generate a DataSource')

    def describe(self):
        return self.summary()

    def __repr__(self):
        return f'<GenerationOperation(options={self.options})>'


class GenerationOptions(OperationOptions):
    """
    Class for options passed to AnalysisOperations
    """
    op_class = GenerationOperation

    def __init__(self, func: Callable[[Any], DataSource], last_modified: Optional[datetime.datetime] = None,
                 out_path: Optional[str] = None, allow_modifying_result: bool = True,
                 result_kwargs: Optional[Dict[str, Any]] = None,
                 **func_kwargs):
        """

        :param func: function which generates the DataSource
        :param last_modified: by default, will always run pipeline when loading generated source, but by passing a
            last_modified can avoid re-running and instead load from location
        :param out_path: location for generated DataSource
        :param func_kwargs: Keyword arguments to pass to the function which generates the DataSource
        :param allow_modifying_result: When DataSources are directly linked to pipelines, loading
            source from pipeline can cause modifications in the pipeline's result source. Set to False
            to ensure it won't be modified (but uses more memory). Setting to False should only be needed
            if multiple sources load from the same pipeline in one session
        """
        self.func = func
        self.func_kwargs = func_kwargs
        self.last_modified = last_modified
        self.out_path = out_path
        self.allow_modifying_result = allow_modifying_result
        self.result_kwargs = result_kwargs



from typing import Union, Sequence, TYPE_CHECKING
if TYPE_CHECKING:
    from datacode.models.source import DataSource
    from datacode.models.pipeline.merge import DataMergePipeline
    from datacode.models.pipeline.generate import DataGeneratorPipeline
    from datacode.models.pipeline.transform import DataTransformationPipeline
    from datacode.models.pipeline.operations.operation import OperationOptions, DataOperation

SourceCreatingPipeline = Union['DataMergePipeline', 'DataGeneratorPipeline', 'DataTransformationPipeline']
DataSourceOrPipeline = Union['DataSource', SourceCreatingPipeline]
DataSourcesOrPipelines = Sequence[DataSourceOrPipeline]
ObjWithLastModified = Union[DataSourceOrPipeline, 'OperationOptions', 'DataOperation']



from typing import List, Optional, Tuple, Any, Dict, Union
import pyexlatex.table as lt
import pandas as pd

StrList = List[str]
StrListOrNone = Optional[StrList]
RegResultSummaryTuple = Tuple[Any, pd.DataFrame]
StrOrNone = Optional[str]
TwoDfTuple = Tuple[pd.DataFrame, pd.DataFrame]
DfDict = Dict[str, pd.DataFrame]
FloatOrNone = Optional[float]
Summary = Union[DfDict, lt.Table]
DfSummaryTuple = Tuple[pd.DataFrame, Summary]


from typing import Optional
import warnings
import math
import pandas as pd
from datacode.psm.score import (
    merge_treated_and_control_calculate_comparison_score,
    EmptyArraysException,
    NotEnoughTimeObservationsException
)
from datacode.psm.dataprep import treated_and_control_df_from_df
from datacode.psm.names import prob_treated_varname
from datacode.psm.typing import StrOrNone, FloatOrNone

def create_matched_df_using_propensity_scores(df: pd.DataFrame,
                                              predict_df: pd.DataFrame,
                                              treated_var: str,
                                              prob_treated_var: StrOrNone = None,
                                              time_var: str = 'Date',
                                              entity_var: str = 'id',
                                              min_matching_pct: FloatOrNone = None,
                                              control_only: bool = False,
                                              num_matches_per_entity: int = 1,
                                              match_with_replacement: bool = False
                                              ) -> pd.DataFrame:
    if prob_treated_var is None:
        prob_treated_var = prob_treated_varname(treated_var)

    treated_df, control_df = treated_and_control_df_from_df(
        predict_df,
        treated_var,
        entity_var,
        keep_vars=[prob_treated_var, entity_var, time_var]
    )

    match_dict = create_matched_id_dict_using_propensity_scores(
        treated_df,
        control_df,
        prob_treated_var,
        time_var=time_var,
        entity_var=entity_var,
        min_matching_pct=min_matching_pct,
        num_matches_per_entity=num_matches_per_entity,
        match_with_replacement=match_with_replacement
    )

    matched_df = create_matched_df_from_match_dict(
        df,
        match_dict,
        entity_var=entity_var,
        control_only=control_only
    )

    return matched_df

def create_matched_df_from_match_dict(df: pd.DataFrame, match_dict: dict, entity_var: str = 'id',
                                      control_only: bool = False) -> pd.DataFrame:
    control_firms_lol = list(match_dict.values())  # list of lists of control firms
    selected_firms = _flatten(control_firms_lol)
    if not control_only:
        selected_treated_firms = list(match_dict.keys())
        selected_firms.extend(selected_treated_firms)  # treated firms

    return df[df[entity_var].isin(selected_firms)]


def create_matched_id_dict_using_propensity_scores(treated_df: pd.DataFrame,
                                                   control_df: pd.DataFrame,
                                                   prob_treated_var: str,
                                                   time_var: str = 'Date',
                                                   entity_var: str = 'id',
                                                   min_matching_pct: FloatOrNone = None,
                                                   num_matches_per_entity: int = 1,
                                                   match_with_replacement: bool = False) -> dict:
    if min_matching_pct is None:
        min_matching_pct = 0

    match_dict = {entity_id: [] for entity_id in treated_df[entity_var].unique()}
    all_matched_firms = []
    for treated_entity_id, treated_entity_df in treated_df.groupby(entity_var):
        comparison_scores = {}
        min_matching_time = math.ceil(len(treated_entity_df) * min_matching_pct)
        for control_entity_id, control_entity_df in control_df.groupby(entity_var):
            if not match_with_replacement and control_entity_id in all_matched_firms:
                # Already matched this control firm, skip
                continue
            try:
                comparison_scores[control_entity_id] = merge_treated_and_control_calculate_comparison_score(
                    treated_entity_df,
                    control_entity_df,
                    prob_treated_var,
                    time_var=time_var,
                    minimum_length=min_matching_time if min_matching_pct != 0 else None
                )
            except EmptyArraysException:
                # No overlapping time observations, cannot be a match
                continue
            except NotEnoughTimeObservationsException:
                # Not enough overlapping time observations, cannot be a match
                continue
        if comparison_scores == {}:
            warnings.warn(f'Could not find another {entity_var} to pair with {treated_entity_id} that has'
                          f' {min_matching_time} overlapping {time_var} obs.')
            continue
        all_scores = list(comparison_scores.values())
        all_scores.sort()
        selected_scores = all_scores[:num_matches_per_entity]
        inverse_dict = {value: key for key, value in comparison_scores.items()}
        for score in selected_scores:
            match_firm = inverse_dict[score]
            match_dict[treated_entity_id].append(match_firm)
            all_matched_firms.append(match_firm)

    print(f'Found {len(all_matched_firms)} matches for {len(match_dict.keys())} entities.')

    return match_dict

def _flatten(list_) -> list:
    return [item for sublist in list_ for item in sublist]


from pandas.testing import assert_frame_equal

from datacode import DataSource
from datacode.models.pipeline.operations.merge import MergeOptions
from tests.pipeline.base import PipelineTest


class TestDataMergePipeline(PipelineTest):

    def test_create_and_run_merge_pipeline_from_sources(self):
        dp = self.create_merge_pipeline()
        dp.execute()

        assert_frame_equal(dp.df, self.expect_merged_1_2)

    def test_create_and_run_merge_pipeline_with_indices(self):
        dp = self.create_merge_pipeline(indexed=True)
        dp.execute()

        assert_frame_equal(dp.df, self.expect_merged_1_2_c_index)

    def test_auto_run_pipeline_by_load_source_with_no_location(self):
        dp = self.create_merge_pipeline()

        ds = DataSource(pipeline=dp, location=self.csv_path_output)
        df = ds.df
        assert_frame_equal(df, self.expect_merged_1_2)

    def test_create_and_run_merge_pipeline_three_sources(self):
        dp = self.create_merge_pipeline(include_indices=(0, 1, 2))
        dp.execute()

        assert_frame_equal(dp.df, self.expect_merged_1_2_3)

    def test_raises_error_for_mismatching_data_sources_merge_options(self):
        mo = MergeOptions([self.merge_var.name])
        dp = self.create_merge_pipeline(include_indices=(0, 1, 2), merge_options_list=[mo])

        with self.assertRaises(ValueError) as cm:
            dp.execute()
            exc = cm.exception
            assert 'must have one fewer merge options than data sources' in str(exc)

    def test_create_nested_pipeline(self):
        dp1 = self.create_merge_pipeline(include_indices=(0, 1))

        self.create_csv_for_3()
        ds3_cols = self.create_columns_for_3()
        ds3 = self.create_source(df=None, location=self.csv_path3, columns=ds3_cols, name='three')

        dp2 = self.create_merge_pipeline(data_sources=[dp1, ds3])
        dp2.execute()

        assert_frame_equal(dp2.df, self.expect_merged_1_2_3)

    def test_create_nested_transformation_pipeline(self):
        dtp = self.create_transformation_pipeline()

        self.create_csv_for_2()
        ds2_cols = self.create_columns_for_2()
        ds2 = self.create_source(df=None, location=self.csv_path2, columns=ds2_cols, name='two')

        dp = self.create_merge_pipeline(data_sources=[dtp, ds2])
        dp.execute()

        assert_frame_equal(dp.df, self.expect_merged_1_transformed_2)

    def test_create_nested_generation_pipeline(self):
        dgp = self.create_generator_pipeline()

        self.create_csv_for_2()
        ds2_cols = self.create_columns_for_2()
        ds2 = self.create_source(df=None, location=self.csv_path2, columns=ds2_cols, name='two')

        dp = self.create_merge_pipeline(data_sources=[dgp, ds2])
        dp.execute()

        assert_frame_equal(dp.df, self.expect_merged_1_generated_2)

    def test_graph(self):
        dp = self.create_merge_pipeline()

        ds = DataSource(pipeline=dp, location=self.csv_path_output)
        df = ds.df

        # TODO [#80]: better tests for graph
        #
        # Currently just checking to make sure they can be generated with no errors.
        # Should also check the contents of the graphs. Also see TestCreateSource.test_graph
        ds.graph
        dp.graph


import pandas as pd
import pd_utils

from pd_utils.pdutils import _to_list_if_str
from datacode.ff.create.model import parse_model
from datacode.ff.fftypes import ListOrStr, DfListListStrTuple, GroupvarNgroupsDict, StrOrInt

def create_ff_portfolios(df: pd.DataFrame, factor_model: StrOrInt=3, byvars: ListOrStr='Year',
                         id_var: str='Ticker', date_var: str='Date',
                         size_var: str=None, value_var: str=None,
                         profitability_var: str=None, investment_var: str=None,
                         custom_groupvar_ngroups_dict: GroupvarNgroupsDict=None,
                         **portfolio_kwargs) -> pd.DataFrame:

    groupvar_ngroups_dict = _get_groupvars_ngroups_dict(
        factor_model=factor_model,
        size_var=size_var,
        value_var=value_var,
        profitability_var=profitability_var,
        investment_var=investment_var,
        custom_groupvars_ngroups_dict=custom_groupvar_ngroups_dict
    )

    return create_portfolios(
        df,
        groupvar_ngroups_dict=groupvar_ngroups_dict,
        byvars=byvars,
        id_var=id_var,
        date_var=date_var,
        **portfolio_kwargs
    )




def create_portfolios(df: pd.DataFrame, groupvar_ngroups_dict: GroupvarNgroupsDict=None, byvars: ListOrStr='Year',
                      id_var: str='Ticker', date_var: str='Date',
                       **portfolio_kwargs) -> pd.DataFrame:

    if groupvar_ngroups_dict is None:
        groupvar_ngroups_dict = {'Market Equity': 2}

    for groupvar, ngroups in groupvar_ngroups_dict.items():
        df = create_portfolio(
            df,
            groupvar=groupvar,
            ngroups=ngroups,
            byvars=byvars,
            id_var=id_var,
            date_var=date_var,
            **portfolio_kwargs
        )

    return df

def create_portfolio(df: pd.DataFrame, groupvar: str='Market Equity', ngroups=2, byvars: ListOrStr='Year',
                    id_var: str='Ticker', date_var: str='Date',
                       **portfolio_kwargs) -> pd.DataFrame:
    portname = _other_groupvar_portname(groupvar)
    ports = _create_portfolio(df, ngroups=ngroups, groupvar=groupvar, byvars=byvars,
                               portvar=portname, id_var=id_var, date_var=date_var, **portfolio_kwargs)

    return ports

def _get_groupvars_ngroups_dict(factor_model: StrOrInt=3,
                                size_var: str=None, value_var: str=None,
                                profitability_var: str=None, investment_var: str=None,
                                custom_groupvars_ngroups_dict: GroupvarNgroupsDict=None) -> GroupvarNgroupsDict:

    factor_model = parse_model(factor_model)

    if factor_model == 'custom':
        return custom_groupvars_ngroups_dict

    groupvars_ngroups_dict = {
        size_var: 2,
        value_var: 3
    }

    if profitability_var is not None:
        groupvars_ngroups_dict.update({
            profitability_var: 3
        })

    if investment_var is not None:
        groupvars_ngroups_dict.update({
            investment_var: 3
        })

    return groupvars_ngroups_dict


def _create_portfolio(df: pd.DataFrame, groupvar: str='Market Equity', ngroups=2, byvars: ListOrStr='Year',
                       portvar: str='portfolio', id_var: str='Ticker', date_var: str='Date',
                       **portfolio_kwargs) -> pd.DataFrame:

    df_for_port_sort, byvars, result_byvars, port_datevar =_create_df_for_port_sort_byvars_result_byvars_port_datevar(
        df,
        groupvar=groupvar,
        byvars=byvars,
        portvar=portvar,
        id_var=id_var,
        date_var=date_var
    )

    # create one portfolio per byvars per id_var
    unique_ports = pd_utils.portfolio(
        df_for_port_sort.dropna(subset=[groupvar]),
        groupvar,
        ngroups=ngroups,
        byvars=byvars,
        portvar=portvar,
        **portfolio_kwargs
    )

    # Merge back to original dataframe
    ports = df.merge(unique_ports[result_byvars + [portvar, port_datevar]], how='left', on=result_byvars)

    return ports


def _create_df_for_port_sort_byvars_result_byvars_port_datevar(df: pd.DataFrame, groupvar: str='Market Equity',
                                                               byvars: ListOrStr='Year', portvar: str='portfolio',
                                                               id_var: str='Ticker',
                                                               date_var: str='Date') -> DfListListStrTuple:

    byvars = _to_list_if_str(byvars)
    result_byvars = byvars + [id_var]

    # create dataframe which has only the beginning of period values for each company
    df_for_port_sort = df.sort_values([id_var, date_var]).groupby(
        result_byvars, as_index=False)[result_byvars + [groupvar, date_var]].first()

    # Rename date variable
    port_datevar = portvar + ' Formation Date'
    assert port_datevar not in df.columns  # don't overwrite existing
    df_for_port_sort.rename(columns={date_var: port_datevar}, inplace=True)


    return df_for_port_sort, byvars, result_byvars, port_datevar


def _other_groupvar_portname(other_groupvar):
    return f'{other_groupvar} Portfolio'


import warnings
from typing import Sequence, Any, Optional, List, Tuple, Dict, Union, Set

from datacode.models.dtypes.base import DataType
from typing_extensions import Protocol
import pandas as pd
from datacode.models.column.column import Column

from datacode.models.variables import Variable
from datacode.models.source import DataSource, NoColumnForVariableException


def combine_sources(data_sources: Sequence[DataSource], rows: bool = True,
                    row_duplicate_vars: Optional[Sequence[Variable]] = None,
                    entity_duplicate_vars: Optional[Sequence[Variable]] = None) -> DataSource:
    if rows:
        return _combine_rows(
            data_sources,
            row_duplicate_vars=row_duplicate_vars,
            entity_duplicate_vars=entity_duplicate_vars
        )

    # Columns
    return _combine_columns(data_sources)


def _combine_columns(data_sources: Sequence[DataSource]) -> DataSource:
    if data_sources[0].index_cols != data_sources[1].index_cols:
        if _column_lists_match_excluding_load_keys(data_sources[0].index_cols, data_sources[1].index_cols):
            _warn_about_mismatching_load_keys(data_sources[0].index_cols, data_sources[1].index_cols)
        else:
            mismatching_message = _mismatching_attributes_of_columns_message(
                data_sources[0].index_cols, data_sources[1].index_cols
            )
            raise ValueError(f'can only combine columns of data sources with overlapping indices. '
                             f'{mismatching_message}')

    new_vars, new_cols = _combine_variables_and_columns(data_sources, allow_overlap=False)
    # new_df = pd.concat([ds.df for ds in data_sources], axis=1)
    new_df = data_sources[0].df.join(data_sources[1].df, how='outer')
    new_source = DataSource(df=new_df, columns=new_cols, load_variables=new_vars)

    return new_source


def _combine_rows(
    data_sources: Sequence[DataSource],
    row_duplicate_vars: Optional[Sequence[Variable]] = None,
    entity_duplicate_vars: Optional[Sequence[Variable]] = None
) -> DataSource:
    if row_duplicate_vars and entity_duplicate_vars:
        raise ValueError('pass at most one of row_duplicate_vars and entity_duplicate_vars')

    reset_index = False
    if not data_sources[0].index_cols and not data_sources[1].index_cols:
        reset_index = True

    new_vars, new_cols = _combine_variables_and_columns(data_sources, allow_overlap=True)

    if not row_duplicate_vars and not entity_duplicate_vars:
        # Simple append
        new_df = pd.concat([ds.df for ds in data_sources])
    elif row_duplicate_vars:
        # Drop duplicates by rows matching row_duplicate_vars, take first source
        new_df = _append_handling_categorical_index(data_sources[0].df, data_sources[1].df)
        _drop_duplicates_by_vars(new_df, row_duplicate_vars)
    else: # entity_duplicate_vars
        # Select entities from the second data source which are not in the first data source, and only
        # add those
        entity_col_names = [var.name for var in entity_duplicate_vars]
        existing_entities = _get_entities(data_sources[0].df, entity_col_names)
        potential_entities = _get_entities(data_sources[1].df, entity_col_names)
        new_entities = list(potential_entities - existing_entities)
        new_rows = _select_df_by_tuples(data_sources[1].df, new_entities, entity_col_names)
        new_df = _append_handling_categorical_index(data_sources[0].df, new_rows)

    if reset_index:
        new_df.reset_index(drop=True, inplace=True)

    # Set output data types
    for col in new_cols:
        col_from_all_sources = []
        for ds in data_sources:
            try:
                col = ds.col_for(var_key=col.variable.unique_key, is_unique_key=True)
            except NoColumnForVariableException:
                # column was not in this source
                continue
            col_from_all_sources.append(col)
        dtypes: Set[DataType] = set([v.dtype for v in col_from_all_sources])
        if len(dtypes) > 1:
            raise NotImplementedError(f'got multiple dtypes {dtypes} for {col} during combine, '
                                      f'need to provide a way for the user to specify output type')
        if len(dtypes) == 0:
            raise ValueError(f'could not extract a data type for {col}')

        dtype = list(dtypes)[0]
        if col.variable.name in new_df.index.names:
            # Variable is in index. Need to handle differently
            if len(new_df.index.names) == 1:
                # Variable is only index
                new_df.index = new_df.index.astype(dtype.index_arg)
                col.series = pd.Series(new_df.index)
            else:
                # Multi-index, need to replace only this level
                idx_pos = new_df.index.names.index(col.variable.name)
                converted = new_df.index.levels[idx_pos].astype(dtype.index_arg)
                new_df.index = new_df.index.set_levels(converted, level=idx_pos)
                col.series = pd.Series(converted)
        else:
            # Variable is in columns, not index
            new_df[col.variable.name] = new_df[col.variable.name].astype(dtype.pd_class())
            col.series = new_df[col.variable.name]

    new_source = DataSource(df=new_df, columns=new_cols, load_variables=new_vars)

    return new_source


def _drop_duplicates_by_vars(df: pd.DataFrame, variables: Sequence[Variable]):
    """
    Drops duplicates on a subset of variables regardless of whether they are in the index

    Notes:
        inplace

    :param df:
    :param variables:
    :return:
    """
    orig_index_cols = df.index.names
    remove_index_name = False
    if orig_index_cols == [None]:
        orig_index_cols = ['index']
        remove_index_name = True

    df.reset_index(inplace=True)
    df.drop_duplicates(subset=[var.name for var in variables], inplace=True)
    df.set_index(orig_index_cols, inplace=True)

    if remove_index_name:
        df.index.name = None


def _append_handling_categorical_index(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
    """
    Resets index, concatenates rows, then sets index back. Only necessary
    because of this pandas issue: https://github.com/pandas-dev/pandas/issues/17629

    Use a regular concat over axis=0 once this issue is resolved.

    :param df:
    :param df2:
    :return:
    """
    if df1.index.names != df2.index.names:
        raise ValueError('must append with identical index')

    orig_index_cols = df1.index.names
    remove_index_name = False
    if orig_index_cols == [None]:
        orig_index_cols = ['index']
        remove_index_name = True

    combined = pd.concat([df.reset_index() for df in [df1, df2]])

    combined.set_index(orig_index_cols, inplace=True)
    if remove_index_name:
        combined.index.name = None
    return combined


def _get_entities(df: pd.DataFrame, entity_col_names: Sequence[str]) -> Set[Sequence[Any]]:
    return set([tuple(r) for r in df.reset_index()[entity_col_names].to_numpy().tolist()])


def _select_df_by_tuples(df: pd.DataFrame, tuples: List[Sequence[Any]],
                         col_names: Sequence[Union[str, int, float]]) -> pd.DataFrame:
    orig_index_cols = df.index.names
    remove_index_name = False
    if orig_index_cols == [None]:
        orig_index_cols = ['index']
        remove_index_name = True

    # Set index to what we are querying by
    df.reset_index(inplace=True)
    df.set_index(col_names, inplace=True)

    # Selection is different when only a single column
    if len(col_names) == 1:
        to_select = [tup[0] for tup in tuples]
    else:
        to_select = tuples

    selected = df.loc[to_select]

    # Set index back to original
    for dataframe in [df, selected]:
        dataframe.reset_index(inplace=True)
        dataframe.set_index(orig_index_cols, inplace=True)
        if remove_index_name:
            dataframe.index.name = None

    return selected


def _combine_variables_and_columns(
    data_sources: Sequence[DataSource],
    allow_overlap: bool = False
) -> Tuple[List[Variable], List[Column]]:
    all_variables = []
    all_columns = []

    for ds in data_sources:
        for var in ds.load_variables:
            if var in all_variables:
                if allow_overlap:
                    continue
                else:
                    if ds.col_for(var) in ds.index_cols:
                        continue
                    raise ValueError(f'variable {var} exists in multiple data sources')
            all_variables.append(var)
            all_columns.append(ds.col_for(var))
    return all_variables, all_columns


def _column_lists_match_excluding_load_keys(cols1: Sequence[Column], cols2: Sequence[Column]) -> bool:
    for col1, col2 in zip(cols1, cols2):
        if not all([
            col1.variable == col2.variable,
            col1.indices == col2.indices,
            col1.applied_transform_keys == col2.applied_transform_keys,
            col1.dtype == col2.dtype
        ]):
            return False

    return True


def _warn_about_mismatching_load_keys(cols1: Sequence[Column], cols2: Sequence[Column]):
    for col1, col2 in zip(cols1, cols2):
        if col1.variable != col2.variable:
            raise ValueError('warn about mismatching must be called only on columns which match other than load_key')
        if col1.load_key != col2.load_key:
            warnings.warn(f'Got both {col1.load_key} and {col2.load_key} as load_key for '
                          f'{col1.variable}, will use {col1.load_key}')


def _mismatching_attributes_of_columns_message(cols1: Sequence[Column], cols2: Sequence[Column]) -> Optional[str]:
    messages = []
    for col1, col2 in zip(cols1, cols2):
        for attr in col1.equal_attrs:
            val1 = getattr(col1, attr)
            val2 = getattr(col2, attr)
            if val1 != val2:
                messages.append(f'Column {col1.load_key} has {val1} for {attr} while {col2.load_key} has {val2}')
    if messages:
        return ', '.join(messages)


class CombineFunction(Protocol):

    def __call__(self, data_sources: Sequence[DataSource], **kwargs: Any) -> DataSource: ...


import pandas as pd
from typing import List, Dict, Tuple, Union, Sequence

StrList = List[str]
FloatSequence = Sequence[float]
AssociatedColDict = Dict[str, StrList]
BoolDict = Dict[str, bool]
DfDict = Dict[str, pd.DataFrame]
TwoDfDictTuple = Tuple[DfDict, DfDict]
TwoDfDictAndDfTuple = Tuple[DfDict, DfDict, pd.DataFrame]
IntOrFloat = Union[int, float]
MinMaxTuple = Tuple[IntOrFloat, IntOrFloat]
MinMaxDict = Dict[str, MinMaxTuple]



import pandas as pd
import pd_utils

from pd_utils.pdutils import _to_list_if_str
from datacode.ff.fftypes import ListOrStr


def portfolio_returns(df: pd.DataFrame, retvar: str='RET', datevar: str='Date', wtvar: str='Market Equity',
                       byvars: ListOrStr=None,
                       portvar: str='Market Equity Portfolio/B/M Portfolio') -> pd.DataFrame:

    byvars = _to_list_if_str(byvars)

    if byvars is not None:
        all_byvars = byvars + [datevar, portvar]
    else:
        all_byvars = [datevar, portvar]

    avgs = pd_utils.averages(
        df,
        retvar,
        byvars=all_byvars,
        wtvar=wtvar,
        count=False
    )

    _get_weighted_averages_from_averages(avgs)

    return avgs

def market_returns(df: pd.DataFrame, retvar: str='RET', datevar: str='Date', wtvar: str='Market Equity',
                   byvars: ListOrStr=None) -> pd.DataFrame:

    byvars = _to_list_if_str(byvars)

    if byvars is not None:
        all_byvars = byvars + [datevar]
    else:
        all_byvars = [datevar]

    avgs = pd_utils.averages(
        df[all_byvars + [retvar, wtvar]],
        retvar,
        byvars=all_byvars,
        wtvar=wtvar,
        count=False
    )

    _get_weighted_averages_from_averages(avgs)
    avgs.rename(columns={retvar: 'MKT'}, inplace=True)

    return avgs


def _get_weighted_averages_from_averages(df: pd.DataFrame) -> None:
    """
    drops EW cols, replaces with VW cols

    Note: inplace
    """
    weighted_average_cols = [col for col in df.columns if col.endswith('_wavg')]
    rename_dict = {col: '_'.join(col.split('_')[:-1]) for col in weighted_average_cols}
    average_cols = [col for col in rename_dict.values()]

    df.drop(average_cols, axis=1, inplace=True)
    df.rename(columns=rename_dict, inplace=True)


from numpy import nan
import pandas as pd
import re

RESULT_PATTERN = re.compile(r'\(([\d.,]+)\)')


def convert_to_stderr_format(value: float, num_decimals: int = 2) -> str:
    if pd.isnull(value):
        return ''
    return f'({value:.{num_decimals}f})'


def parse_stderr_value(value: str) -> float:
    match = RESULT_PATTERN.fullmatch(value)
    if not match:
        return nan

    result = match.group(1)
    return float(result)



import pandas as pd
from dateutil.relativedelta import relativedelta

from datacode.typing import StrList

def expand_entity_date_selections(full_df: pd.DataFrame, selections_df: pd.DataFrame, cols: StrList = None,
                                  num_firms: int = 3, expand_months: int = 3,
                                  entity_id_col: str = 'TICKER',
                                  date_col: str = 'Date',
                                  begin_datevar: str = 'Begin Date', end_datevar: str = 'End Date',
                                  ) -> pd.DataFrame:

    entity_date_df = _firm_date_range_df_from_df(
        selections_df,
        num_firms=num_firms,
        firm_id_col=entity_id_col,
        date_col=date_col,
        begin_datevar=begin_datevar,
        end_datevar=end_datevar
    )
    _expand_date_df(
        entity_date_df,
        expand_months=expand_months,
        begin_datevar=begin_datevar,
        end_datevar=end_datevar
    )
    entity_df = _select_orig_df_from_date_df(
        full_df,
        entity_date_df,
        firm_id_col=entity_id_col,
        date_col=date_col,
        begin_datevar=begin_datevar,
        end_datevar=end_datevar,
        associated_cols=cols
    )

    return entity_df



def _firm_date_range_df_from_df(df: pd.DataFrame, num_firms: int = 3, firm_id_col: str = 'TICKER',
                                date_col: str = 'Date',
                                begin_datevar: str = 'Begin Date', end_datevar: str = 'End Date') -> pd.DataFrame:
    firm_date_vars = [firm_id_col, date_col]

    firms = df[firm_id_col].unique()[:num_firms]
    firm_dates = df[df[firm_id_col].isin(firms)][firm_date_vars].sort_values(firm_date_vars)
    earliest_dates = firm_dates.groupby(firm_id_col).min()
    earliest_dates.rename(columns={date_col: begin_datevar}, inplace=True)
    latest_dates = firm_dates.groupby(firm_id_col).max()
    latest_dates.rename(columns={date_col: end_datevar}, inplace=True)
    return earliest_dates.join(latest_dates)


def _expand_date_df(date_df: pd.DataFrame, expand_months: int = 3,
                    begin_datevar: str = 'Begin Date', end_datevar: str = 'End Date'):
    """
    Note: inplace
    """
    date_df[begin_datevar] = date_df[begin_datevar].apply(lambda x: x - relativedelta(months=expand_months))
    date_df[end_datevar] = date_df[end_datevar].apply(lambda x: x + relativedelta(months=expand_months))


def _select_orig_df_from_date_df(df: pd.DataFrame, date_df: pd.DataFrame,
                                 firm_id_col: str = 'TICKER',
                                 date_col: str = 'Date',
                                 begin_datevar: str = 'Begin Date', end_datevar: str = 'End Date',
                                 associated_cols: StrList = None) -> pd.DataFrame:
    if associated_cols == None:
        associated_cols = [firm_id_col, date_col]

    out_df = pd.DataFrame()
    for firm_id, date_series in date_df.iterrows():
        selected_firm_df = df.loc[
            (df[firm_id_col] == firm_id) &
            (df[date_col] >= date_series[begin_datevar]) &
            (df[date_col] <= date_series[end_datevar]),
            associated_cols
        ]
        out_df = out_df.append(selected_firm_df)

    return out_df


from functools import partial
import pandas as pd

from datacode.portfolio.name import(
    _is_differenced_port_name,
    _ports_from_differenced_port_name
)

def sort_portfolio_index(df, inplace=False):
    if not inplace:
        df = df.copy()

    num_ports = len(df.index.unique())
    _port_sort_key = partial(_port_sort_value, num_ports=num_ports)
    index_name = df.index.name
    index_access_name = 'index' if index_name is None else index_name
    df.reset_index(inplace=True)
    df['_sort_value'] = df[index_access_name].apply(_port_sort_key).astype(int)
    df.sort_values('_sort_value', inplace=True)
    df.set_index(index_access_name, inplace=True)
    df.drop('_sort_value', axis=1, inplace=True)
    df.index.name = index_name

    if not inplace:
        return df

def sort_portfolio_columns(df):
    """
    Note: creates a copy, not inplace
    """
    return sort_portfolio_index(df.T).T

def sort_portfolio_columns_and_index(df: pd.DataFrame) -> pd.DataFrame:
    """
    Note: creates a copy, not inplace
    """
    df = sort_portfolio_index(df, inplace=False)
    return sort_portfolio_columns(df)

def _port_sort_value(int_or_str, num_ports=3):
    if _is_int(int_or_str):
        return int(int_or_str)

    if not isinstance(int_or_str, str):
        raise ValueError(f'must pass str or int, got {int_or_str} of type {type(int_or_str)}')

    name_dict = {
        'high': num_ports,
        'mid': 2, # mid only appears when there are three ports
        'low': 1,
        'zero': 0
    }
    name = int_or_str.lower().strip()

    if name in name_dict:
        return name_dict[name]

    # If this is a differenced portfolio, use the sum of the sort values of the individual, plus the max port.
    # This way differenced portfolios will always come after other portfolios, but then will be ordered consistently
    if _is_differenced_port_name(name):
        long_port, short_port = _ports_from_differenced_port_name(name)
        long_port_sort_value = _port_sort_value(long_port, num_ports=num_ports)
        short_port_sort_value = _port_sort_value(short_port, num_ports=num_ports)
        return long_port_sort_value + short_port_sort_value + num_ports # num ports to ensure after individual ports


    raise ValueError(f'passed str was not an int or special port name high, low, or zero. got {int_or_str}')


def _is_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


import pandas as pd
from datacode.psm.typing import StrListOrNone, TwoDfTuple

def treated_and_control_df_from_df(df: pd.DataFrame, treated_var: str, entity_var: str,
                                   keep_vars: StrListOrNone = None) -> TwoDfTuple:
    rows_with_treatment = df[df[treated_var] == 1]
    # Cannot simply take where treatment is zero as control, as some entities may be treated for only a portion of the
    # sample. If this is the case, would end up with the entity in both dfs. Must set control as those entities
    # that are never treated
    treated_entities = rows_with_treatment[entity_var].unique()
    treated_df = df[df[entity_var].isin(treated_entities)]
    control_df = df[~df[entity_var].isin(treated_entities)]


    if keep_vars is not None:
        treated_df = treated_df[keep_vars]
        control_df = control_df[keep_vars]

    return treated_df, control_df




from typing import Dict, Sequence, List

import pyexlatex as pl

from datacode.models.variables import Variable


def model_eqs(structural_dict: Dict[Variable, Sequence[Variable]],
              measurement_dict: Dict[Variable, Sequence[Variable]],
              var_corr_groups: Sequence[Sequence[Variable]]) -> List[pl.Equation]:
    all_eqs = []
    for y, x_vars in structural_dict.items():
        all_vars = [y, *x_vars]
        all_eqs.append(_vars_to_eq(all_vars, operator='='))
    for y, x_vars in measurement_dict.items():
        all_vars = [y, *x_vars]
        all_eqs.append(_vars_to_eq(all_vars, operator='='))
    for corr_group in var_corr_groups:
        all_eqs.append(_vars_to_eq(corr_group, operator='~~'))
    valid_eqs = [eq for eq in all_eqs if eq]
    return valid_eqs


def _vars_to_eq(var_seq: Sequence[Variable], operator: str = '=', inline: bool = False):
    if len(var_seq) < 2 or (len(var_seq) == 2 and var_seq[0] == var_seq[1]):
        return None

    lhs = var_seq[0].symbol
    rhs = ' + '.join([str(var.symbol) for var in var_seq[1:]])

    if operator == '=':
        eq_str = f'{lhs} = {rhs}'
    elif operator == '~~':
        eq_str = rf'{lhs} \sim {rhs}'
    else:
        raise NotImplementedError(f'operator {operator} not supported')
    return pl.Equation(str_eq=eq_str, inline=inline)


from typing import Optional, Sequence, Union, List

import pandas as pd
from mixins import ReprMixin
from mixins.attrequals import EqOnAttrsMixin

from datacode.models.column.index import ColumnIndex
from datacode.models.dtypes.base import DataType
from datacode.models.dtypes.convert import convert_str_to_data_type_if_necessary
from datacode.models.variables import Variable
from datacode.models.transform.transform import Transform


class Column(EqOnAttrsMixin, ReprMixin):
    equal_attrs = [
        'variable',
        'indices',
        'load_key',
        'applied_transform_keys',
        'dtype',
    ]
    repr_cols = ['variable', 'load_key', 'indices', 'applied_transform_keys', 'dtype']

    def __init__(self, variable: Variable, load_key: Optional[str] = None,
                 indices: Optional[Sequence[ColumnIndex]] = None,
                 applied_transform_keys: Optional[Sequence[str]] = None,
                 dtype: Optional[Union[str, DataType]] = None,
                 series: Optional[pd.Series] = None):
        if applied_transform_keys is None:
            applied_transform_keys = []

        dtype = convert_str_to_data_type_if_necessary(dtype)

        if dtype is None:
            dtype = variable.dtype

        self.load_key = load_key
        self.variable = variable
        self.indices = indices
        self.applied_transform_keys = applied_transform_keys
        self.dtype = dtype
        self.series = series

    @property
    def index_vars(self) -> List[Variable]:
        if self.indices is None:
            return []

        index_vars = []
        for col_idx in self.indices:
            for var in col_idx.variables:
                if var not in index_vars:
                    index_vars.append(var)
        return index_vars

    def _add_applied_transform(self, transform: Transform, skip_variable: bool = False):
        self.applied_transform_keys.append(transform.key)
        if not skip_variable:
            self.variable._add_applied_transform(transform)



from typing import Union, Type

import pandas as pd

from datacode.models.dtypes.base import DataType


class StringType(DataType):
    names = ('str', 'string', 'character', 'str_')

    def __init__(self, categorical: bool = False, ordered: bool = False):
        # TODO [#23]: add all string methods as available transforms
        super().__init__(
            str,
            pd_class=pd.StringDtype,
            categorical=categorical,
            ordered=ordered
        )

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        dtype = dtype.lower()
        if dtype not in cls.names:
            raise ValueError(f'Dtype {dtype} does not match valid names for {cls.__name__}: {cls.names}')
        return cls(
            categorical=categorical,
            ordered=ordered
        )

    @property
    def read_file_arg(self) -> Union[Type, str]:
        if self.categorical:
            return 'category'
        return 'string'

    @property
    def index_arg(self) -> Union[Type, str]:
        return self.pd_class



import uuid
from copy import deepcopy
import pandas as pd
from functools import partial
import os
import warnings
import datetime
from typing import List, Optional, Any, Dict, Sequence, Type

from graphviz import Digraph
from mixins import ReprMixin

from datacode.graph.base import GraphObject
from datacode.graph.edge import Edge
from datacode.graph.node import Node
from datacode.models.outputter import DataOutputter
from datacode.models.types import SourceCreatingPipeline
from datacode.summarize import describe_df

from datacode.models.variables.variable import Variable
from datacode.models.column.column import Column
from datacode.models.loader import DataLoader
import datacode.hooks as hooks


class DataSource(ReprMixin):
    copy_keys = [
        'location',
        'name',
        'tags',
        'loader_class',
        'outputter_class',
        'pipeline',
        'columns',
        'load_variables',
        'read_file_kwargs',
        'data_outputter_kwargs',
        'optimize_size',
    ]
    update_keys = copy_keys + [
        '_orig_columns',
        '_columns_for_calculate',
        '_orig_load_variables',
        '_vars_for_calculate',
    ]
    repr_cols = ['name', 'pipeline', 'columns', 'load_variables']

    def __init__(self, location: Optional[str] = None, df: Optional[pd.DataFrame] = None,
                 pipeline: Optional[SourceCreatingPipeline] = None,
                 columns: Optional[Sequence[Column]] = None,
                 load_variables: Optional[Sequence[Variable]] = None,
                 name: Optional[str] = None, tags: Optional[List[str]] = None,
                 loader_class: Optional[Type[DataLoader]] = None, read_file_kwargs: Optional[Dict[str, Any]] = None,
                 outputter_class: Optional[Type[DataOutputter]] = None,
                 data_outputter_kwargs: Optional[Dict[str, Any]] = None,
                 optimize_size: bool = False):

        if read_file_kwargs is None:
            read_file_kwargs = {}
        if data_outputter_kwargs is None:
            data_outputter_kwargs = {}
        if loader_class is None:
            loader_class = DataLoader
        if outputter_class is None:
            outputter_class = DataOutputter
        if load_variables is None and columns is not None:
            load_variables = [col.variable for col in columns]
        if columns is not None and not isinstance(columns, list):
            columns = list(columns)

        # Handle setup for loading columns needed only for calculations
        extra_cols_for_calcs = []
        extra_vars_for_calcs = []
        if load_variables is not None and columns is not None:
            vars_for_calculate = []
            for var in load_variables:
                if var.calculation is not None:
                    for var_for_calc in var.calculation.variables:
                        if var_for_calc not in vars_for_calculate:
                            vars_for_calculate.append(var_for_calc)
            for var in vars_for_calculate:
                all_vars = load_variables + extra_vars_for_calcs
                current_var_keys = [load_var.key for load_var in all_vars]
                if var not in all_vars:
                    # TODO [#40]: don't trigger extra columns when the extra columns are just the untransformed columns
                    #
                    # We are adding extra columns here for calculated variables which require variables not
                    # included in `load_variables`. Currently, it will load extra variables even if
                    # the calculation could just be done before variable transforms. For example, the
                    # test `TestLoadSource.test_load_with_calculate_on_transformed_before_transform` should be able
                    # to complete without adding any extra columns
                    if var.key in current_var_keys:
                        # This is a variable that is already being loaded but with a different transformation
                        # Need to create an extra column for it for later loading handling
                        existing_col = [col for col in columns if col.variable.key == var.key][0]
                        new_col = deepcopy(existing_col)
                        new_col.variable = var
                        extra_cols_for_calcs.append(new_col)
                    extra_vars_for_calcs.append(var)

        if columns is not None:
            all_columns = columns + extra_cols_for_calcs
        else:
            all_columns = columns

        if load_variables is not None:
            all_load_vars = load_variables + extra_vars_for_calcs
        else:
            all_load_vars = load_variables

        self._check_inputs(location, df)
        self.location = location
        self.name = name
        self.tags = tags # TODO: better handling for tags
        self.loader_class = loader_class
        self.outputter_class = outputter_class
        self.pipeline = pipeline
        self._orig_columns: Optional[List[Column]] = columns
        self._columns_for_calculate = extra_cols_for_calcs
        self.columns: Optional[List[Column]] = all_columns
        self._orig_load_variables = load_variables
        self._vars_for_calculate = extra_vars_for_calcs
        self.load_variables = all_load_vars
        self.read_file_kwargs = read_file_kwargs
        self.data_outputter_kwargs = data_outputter_kwargs
        self.optimize_size = optimize_size
        self.primary_node = Node(self.name)
        self._df = df

        self._validate()

    def _validate(self):
        self._validate_load_variables()

    def _validate_load_variables(self):
        if self.load_variables is None:
            return
        # Ensure uniqueness of loaded variables' names
        var_names = [var.name for var in self.load_variables]
        existing_names = []
        for name in var_names:
            if name in existing_names:
                raise ValueError(f'variable name {name} repeated in load variables')
            existing_names.append(name)


    @property
    def df(self):
        if self._df is None:
            self._df = self._load()
        return self._df

    @df.setter
    def df(self, df):
        self._df = df
        self.refresh_columns_series()

    @property
    def last_modified(self) -> Optional[datetime.datetime]:
        if self.location is None or not os.path.exists(self.location):
            # No location. Will trigger pipeline instead
            return None

        return datetime.datetime.fromtimestamp(os.path.getmtime(self.location))

    def _load(self):
        hooks.on_begin_load_source(self)
        if not hasattr(self, 'data_loader'):
            self._set_data_loader(self.loader_class, pipeline=self.pipeline, **self.read_file_kwargs)
        df = self.data_loader()
        df = hooks.on_end_load_source(self, df)
        return df

    def output(self, **data_outputter_kwargs):
        config_dict = deepcopy(self.data_outputter_kwargs)
        config_dict.update(**data_outputter_kwargs)
        outputter = self.outputter_class(self, **config_dict)
        outputter.output()

    def _check_inputs(self, filepath, df):
        pass
        # assert not (filepath is None) and (df is None)

    def _set_data_loader(self, data_loader_class: Type[DataLoader], pipeline: SourceCreatingPipeline = None,
                         **read_file_kwargs):
        run_pipeline = False
        if pipeline is not None:
            # if a source in the pipeline to create this data source was modified more recently than this data source
            # note: if there is no location, will always enter the next block, as last modified time will set
            # to a long time ago
            if (
                    # no existing location for this source, must use pipeline
                    self.last_modified is None or
                    # not able to determine when pipeline sources were modified, must always run pipeline
                    pipeline.last_modified is None or
                    # pipeline sources were modified more recently than this source, run pipeline
                    pipeline.last_modified > self.last_modified
            ):
                # a prior source used to construct this data source has changed. need to re run pipeline
                run_pipeline = True
                if pipeline.last_modified is None:
                    warnings.warn(f"""
                    Was not able to determine last modified of pipeline {pipeline.name}.
                    Will always run pipeline due to this. Consider manually setting last_modified when creating
                    the pipeline.
                    """.strip())
                elif self.last_modified is None:
                    warnings.warn(f"""
                   Was not able to determine last modified of source {self.name}.
                   Will run pipeline due to this. This is due to no file currently existing for this source.
                   """.strip())
                else:
                    recent_obj = pipeline.obj_last_modified
                    try:
                        recent_obj_name = recent_obj.name
                    except AttributeError:
                        # Must be Operation, get name from pipeline instead
                        recent_obj_name = pipeline.name
                    warnings.warn(f'''{recent_obj_name} was modified at {recent_obj.last_modified}.
    
                    this data source {self.name} was modified at {self.last_modified}.
    
                    to get new changes, will load this data source through pipeline rather than from file.''')

            # otherwise, don't need to worry about pipeline, continue handling

        loader = data_loader_class(self, read_file_kwargs, self.optimize_size)

        # If necessary, run pipeline before loading
        # Still necessary to use loader as may be transforming the data
        if run_pipeline:
            def run_pipeline_then_load(pipeline: SourceCreatingPipeline):
                pipeline.execute() # outputs to file
                result = loader.load_from_existing_source(
                    pipeline.result,
                    preserve_original=not pipeline.allow_modifying_result
                )
                return result
            self.data_loader = partial(run_pipeline_then_load, self.pipeline)
        else:
            self.data_loader = loader.load_from_location

    def update_from_source(self, other: 'DataSource', exclude_attrs: Optional[Sequence[str]] = tuple(),
                           include_attrs: Optional[Sequence[str]] = tuple()):
        """
        Updates attributes of this DataSource with another DataSources attributes

        :param other:
        :param exclude_attrs: Any attributes to exclude when updating
        :param include_attrs: Defaults to DataSource.update_keys + ['_df'] but can manually select attributes
        :return:
        """
        if not include_attrs:
            include_attrs = self.update_keys + ['_df']

        for attr in include_attrs:
            if attr not in exclude_attrs:
                other_value = getattr(other, attr)
                setattr(self, attr, other_value)

    def copy(self, **kwargs):
        if not kwargs:
            return deepcopy(self)

        config_dict = {attr: deepcopy(getattr(self, attr)) for attr in self.copy_keys}

        # Handle df only if not passed as do not want load df unnecessarily
        if 'df' not in kwargs:
            config_dict['df'] = self.df

        config_dict.update(kwargs)

        klass = type(self)
        return klass(**config_dict)

    def untransformed_col_for(self, variable: Variable) -> Column:
        possible_cols = [col for col in self.columns if col.variable.key == variable.key]
        var_applied_transform_keys = [transform.key for transform in variable.applied_transforms]
        for col in possible_cols:
            if not col.applied_transform_keys:
                # Matches key and no transformations, this is the correct col
                return col
            # Check for subset of transformations
            for i, var_transform_key in enumerate(var_applied_transform_keys):
                if i + 1 >= len(col.applied_transform_keys):
                    # More transforms applied to var than in col, must be the matching col
                    return col
                if col.applied_transform_keys[i] != var_applied_transform_keys[i]:
                    # Mismatching transform between column and variable, must be incorrect column
                    break
        raise NoColumnForVariableException(f'could not find untransformed col for {variable} in {self.columns}')

    def col_for(self, variable: Optional[Variable] = None, var_key: Optional[str] = None,
                orig_only: bool = False, for_calculate_only: bool = False,
                is_unique_key: bool = False) -> Column:
        try:
            # Prefer exact match, need to try this because there may be multiple columns with identical
            # variables besides the applied transforms
            col = self._col_for(
                variable, var_key, orig_only=orig_only,
                for_calculate_only=for_calculate_only, is_unique_key=is_unique_key
            )
        except NoColumnForVariableException:
            # Fall back to matching only on variable key as it may be that it is the correct column but
            # the transforms have not been applied yet
            col = self._col_for(
                variable, var_key, match_key_only=True,
                orig_only=orig_only, for_calculate_only=for_calculate_only,
                is_unique_key=is_unique_key,
            )
        return col

    def _col_for(self, variable: Optional[Variable] = None, var_key: Optional[str] = None,
                 match_key_only: bool = False, orig_only: bool = False, for_calculate_only: bool = False,
                 is_unique_key: bool = False) -> Column:
        if orig_only and for_calculate_only:
            raise ValueError('pass only one of orig_only and for_calculate_only')
        if variable is None and var_key is None:
            raise ValueError('must pass variable or variable key')

        if orig_only:
            col_list = self._orig_columns
        elif for_calculate_only:
            col_list = self._columns_for_calculate
        else:
            col_list = self.columns

        if col_list is None:
            raise NoColumnForVariableException(f'no columns in {self.name}')

        if is_unique_key:
            key_attr = 'unique_key'
        else:
            key_attr = 'key'

        if variable is None:
            possible_cols = [col for col in self.columns if getattr(col.variable, key_attr) == var_key]
            if len(possible_cols) == 0:
                raise NoColumnForVariableException(f'cannot look up col for key {var_key} as no columns match')
            if len(possible_cols) > 1:
                raise ValueError(f'cannot look up col for key {var_key} as multiple columns match: {possible_cols}')
            variable = possible_cols[0].variable
        for col in col_list:
            if match_key_only:
                if getattr(col.variable, key_attr) == getattr(variable, key_attr):
                    return col
            else:
                if col.variable == variable:
                    return col
        raise NoColumnForVariableException(f'could not find column matching {variable} in {self.columns}')

    def col_key_for(self, variable: Optional[Variable] = None, var_key: Optional[str] = None,
                    orig_only: bool = False, for_calculate_only: bool = False) -> Column:
        col = self.col_for(variable, var_key, orig_only=orig_only, for_calculate_only=for_calculate_only)
        return col.load_key

    @property
    def col_var_keys(self) -> List[str]:
        return [col.variable.key for col in self.columns]

    @property
    def col_load_keys(self) -> List[str]:
        return [col.load_key for col in self.columns]

    @property
    def load_var_keys(self) -> List[str]:
        if self.load_variables is None:
            return []

        return [var.key for var in self.load_variables]

    def refresh_columns_series(self):
        if self.columns is None:
            return
        for col in self.columns:
            if col.variable not in self.load_variables:
                continue
            if col.variable.name not in list(self._df.columns) + list(self._df.index.names):
                col.series = None
                continue
            series = self.get_series_for(var=col.variable)
            col.series = series

    def get_var_by_key(self, key: str, is_unique_key: bool = False) -> Variable:
        if is_unique_key:
            attr = 'unique_key'
        else:
            attr = 'key'

        collected_variables = []
        for var in self.load_variables:
            match_key = getattr(var, attr)
            if key == match_key:
                collected_variables.append(var)

        matching_str = f'matching {"unique " if is_unique_key else ""}key {key}'
        if len(collected_variables) == 0:
            raise ValueError(f'could not find variable {matching_str}')
        if len(collected_variables) > 1:
            raise ValueError(f'got multiple variables {matching_str}')

        selected_var = collected_variables[0]
        return selected_var

    def get_series_for(self, var_name: Optional[str] = None, var: Optional[Variable] = None,
                       col: Optional[Column] = None, df: Optional[pd.DataFrame] = None) -> pd.Series:
        """
        Extracts series for a variable or column, regardless of whether it is a column or index

        :param var_name:
        :param var:
        :param col:
        :param df: Will use source.df by default, but can also pass a custom df to use
        :return:
        """
        # Validate inputs
        conditions = [
            var_name is not None,
            var is not None,
            col is not None
        ]
        num_passed = len([cond for cond in conditions if cond])

        if num_passed == 0:
            raise ValueError('must pass one of var_name, var, or col to get series')
        elif num_passed > 1:
            raise ValueError('must pass at most one of var_name, var, or col to get series')

        # Main logic
        if var is not None:
            var_name = var.name
        if col is not None:
            var_name = col.variable.name
        if df is None:
            df = self.df

        if var_name in df.index.names:
            # Need to get from index and convert to series
            return pd.Series(df.index.get_level_values(var_name))
        else:
            # Regular column, just look it up normally
            return df[var_name]

    @property
    def index_cols(self) -> List[Column]:
        if self.columns is None:
            return []

        index_vars = self.index_vars

        index_cols = []
        for col in self.columns:
            if col.variable in index_vars:
                index_cols.append(col)

        return index_cols

    @property
    def index_vars(self) -> List[Variable]:
        if self.columns is None:
            return []

        index_vars = []
        for col in self.columns:
            if col.indices:
                for col_idx in col.indices:
                    for var in col_idx.variables:
                        if var not in index_vars:
                            index_vars.append(var)

        # Sort according to order passed in load_variables
        index_vars.sort(key=lambda x: self.load_var_keys.index(x.key))

        return index_vars

    @property
    def index_var_names(self) -> List[str]:
        index_vars = self.index_vars
        return [var.name for var in index_vars]

    @property
    def loaded_columns(self) -> Optional[List[Column]]:
        if self.columns is None:
            return None
        cols = []
        for var in self.load_variables:
            col = self.col_for(var)
            cols.append(col)
        return cols

    def describe(self):
        # TODO [#48]: use columns, variables, indices, etc. in describe
        return describe_df(self.df)

    def _create_series_in_df_for_calculation(self, df: pd.DataFrame, col: Column):
        new_key = str(uuid.uuid4())  # temporary key for this variable
        # should get column which already has data for this variable
        existing_col = self.col_for(col.variable)
        df[new_key] = deepcopy(df[existing_col.load_key])
        col.load_key = new_key

    def _duplicate_column_for_calculation(self, df: pd.DataFrame, orig_var: Variable, new_var: Variable,
                                          pre_rename: bool = True):
        # should get column which already has data for this variable
        existing_col = self.col_for(orig_var)

        if pre_rename:
            existing_var_name = existing_col.load_key
        else:
            existing_var_name = orig_var.name

        col = deepcopy(existing_col)
        col.variable = new_var

        if pre_rename:
            new_key = str(uuid.uuid4())  # temporary key for this variable
            df[new_key] = deepcopy(df[existing_var_name])
            col.load_key = new_key
        else:
            df[new_var.name] = deepcopy(df[existing_var_name])

        self.columns.append(col)

    @property
    def _graph_contents(self) -> Sequence[GraphObject]:
        elems = [self.primary_node]
        if self.pipeline is not None:
            elems.extend(self.pipeline._graph_contents)
            elems.append(Edge(self.pipeline.primary_node, self.primary_node))
        return elems

    @property
    def graph(self) -> Digraph:
        elems = self._graph_contents
        graph = Digraph(self.name)
        for elem in elems:
            elem.add_to_graph(graph)
        return graph


class NoColumnForVariableException(Exception):
    pass



from typing import Optional

from datacode.models.pipeline.operations.transform import TransformOptions
from datacode.models.source import DataSource

from datacode.models.pipeline.base import DataPipeline


class DataTransformationPipeline(DataPipeline):
    """
    A DataPipeline which creates a DataSource directly from a single other DataSource
    """

    def __init__(self, data_source: DataSource, options: TransformOptions, name: Optional[str] = None):
        super().__init__(
            [data_source],
            [options],
            name=name
        )

    @property
    def data_source(self) -> DataSource:
        return self.data_sources[0]

    @data_source.setter
    def data_source(self, source: DataSource):
        self.data_sources[0] = source



from typing import TYPE_CHECKING

from sympy import Symbol

from datacode.models.transform.transform import Transform
from datacode.models.transform.applied import AppliedTransform

if TYPE_CHECKING:
    from datacode.models.variables.variable import Variable
    from datacode.models.column.column import Column
    from datacode.models.source import DataSource


def create_changes_name_func(name: str, **kwargs) -> str:
    return name + ' Change'


def create_changes_symbol_func(sym: Symbol, **kwargs) -> Symbol:
    sym_str = str(sym)
    new_sym_str = r'\delta ' + sym_str
    sym = Symbol(new_sym_str)
    return sym


def create_changes_data_func(col: 'Column', variable: 'Variable', source: 'DataSource', **kwargs) -> 'DataSource':
    from datacode.models.transform.specific.lag import lag_transform

    applied_lag_transform = AppliedTransform.from_transform(lag_transform, **kwargs)
    source_for_lag = source.copy(
        df=source.df[[variable.name]]
    )
    source_with_lag = applied_lag_transform.apply_to_source(
        source_for_lag,
        preserve_original=False,
        subset=[variable]
    )
    source.df[variable.name] = source.df[variable.name] - source_with_lag.df[variable.lag(**kwargs).name]

    return source


change_transform = Transform(
    'change',
    name_func=create_changes_name_func,
    data_func=create_changes_data_func,
    symbol_func=create_changes_symbol_func,
    data_func_target='source'
)



from typing import Dict, Union, Tuple, Optional, List, Sequence
import pandas as pd

from pyexlatex import Document
from pyexlatex.table import Table
from pyexlatex.figure import Figure

SimpleDfDict = Dict[str, pd.DataFrame]
DfOrDfDict = Union[SimpleDfDict, pd.DataFrame]
DfDict = Dict[str, DfOrDfDict]
DfDictOrNone = Union[DfDict, None]
DfOrSeries = Union[pd.DataFrame, pd.Series]
DfTuple = Tuple[DfOrSeries, DfOrSeries]
DfTupleOrNone = Optional[DfTuple]
FloatList = List[float]
FloatOrNone = Union[float, None]
FloatSequence = Sequence[float]
FloatSequenceOrNone = Union[FloatSequence, None]
StrList = List[str]
StrOrNone = Union[str, None]
DocumentOrTable = Union[Document, Table]
Tables = List[Table]
DocumentOrTables = Union[Document, Tables]
DocumentOrTablesOrNone = Union[DocumentOrTables, None]
LatexObj = Union[Table, Figure]
LatexObjs = List[LatexObj]
DocumentOrLatexObjs = Union[Document, LatexObjs]
IntSequence = Sequence[int]
IntSequenceOrNone = Union[IntSequence, None]
IntOrNone = Union[int, None]


import pandas as pd
import pyexlatex.table as lt

def obs_count_and_missing_data_table(df: pd.DataFrame, col_with_missings: str,
                                     missing_display_str: str='Missing', pct_format_str: str='.1f') -> lt.DataTable:

    obs_pct_df = _obs_count_and_missing_pct_df(
        df,
        col_with_missings,
        missing_display_str=missing_display_str,
        pct_format_str=pct_format_str
    )

    obs_pct_dt = lt.DataTable.from_df(
        obs_pct_df,
        extra_header='Obs'
    )

    return obs_pct_dt

def _obs_count_and_missing_pct_df(df: pd.DataFrame, col_with_missings: str,
                                  missing_display_str: str='Missing', pct_format_str: str='.1f') -> pd.DataFrame:
    obs_count = len(df)
    non_miss_count = len(df.dropna(subset=[col_with_missings]))
    return pd.DataFrame(
        data=[
            (
                obs_count,
                f'{(1 - (non_miss_count/obs_count)) * 100:{pct_format_str}}'
            )
        ],
        columns=['#', f'% {missing_display_str}']
    )


from datacode import AnalysisOptions
from tests.pipeline.base import PipelineTest, analysis_from_source


class TestDataAnalysisPipeline(PipelineTest):

    def test_create_and_run_analysis_pipeline_from_source(self):
        dap = self.create_analysis_pipeline()
        dap.execute()

        assert dap.result.result == self.ds_one_analysis_result

    def test_analysis_pipeline_output(self):
        options = AnalysisOptions(analysis_from_source, out_path=self.csv_path_output)
        dap = self.create_analysis_pipeline(options=options)
        dap.execute()

        with open(self.csv_path_output, 'r') as f:
            result_from_file = int(f.read())

        assert result_from_file == self.ds_one_analysis_result

    def test_create_and_run_analysis_pipeline_from_merge_pipeline(self):
        dmp = self.create_merge_pipeline()
        dap = self.create_analysis_pipeline(source=dmp)
        dap.execute()

        assert dap.result.result == self.ds_one_and_two_analysis_result

    def test_create_and_run_analysis_pipeline_from_transformation_pipeline(self):
        dtp = self.create_transformation_pipeline()
        dap = self.create_analysis_pipeline(source=dtp)
        dap.execute()

        assert dap.result.result == self.ds_one_transformed_analysis_result

    def test_create_and_run_analysis_pipeline_from_generation_pipeline(self):
        dgp = self.create_generator_pipeline()
        dap = self.create_analysis_pipeline(source=dgp)
        dap.execute()

        assert dap.result.result == self.ds_one_generated_analysis_result



from datacode import StringType, FloatType, IntType, BooleanType, TimedeltaType, DatetimeType, ObjectType


def test_dtype_eq():
    assert StringType() == StringType()


def test_is_numeric():
    assert FloatType().is_numeric == IntType().is_numeric == True
    assert StringType().is_numeric == BooleanType().is_numeric == TimedeltaType().is_numeric == \
           DatetimeType().is_numeric == ObjectType().is_numeric == False



import pandas as pd
from datacode.typing import StrList

from datacode.typing import IntOrNone
from datacode.summarize.subset.missing.detail.textfuncs import (
    missing_more_than_str,
    missing_more_than_pct_str,
    missing_tolerance_count_str,
    id_count_str
)


def by_id_pct_long_df(df: pd.DataFrame, byvars: StrList, id_var: str,
                      count_with_missings_var: str, missing_tolerance: IntOrNone = 0,
                      missing_quantile: IntOrNone = None,
                      missing_display_str: str = 'Missing') -> pd.DataFrame:

    by_id_var = _by_id_count_long_df(
        df,
        byvars,
        id_var,
        count_with_missings_var,
        missing_tolerance=missing_tolerance,
        missing_quantile=missing_quantile,
        missing_display_str=missing_display_str
    )

    name_args = (missing_tolerance, missing_quantile, missing_display_str, id_var)
    missing_more_than_name = missing_more_than_str(*name_args)
    missing_tolerance_count_name = missing_tolerance_count_str(*name_args)
    id_count_name = id_count_str(id_var)

    by_id_var[id_count_name] = by_id_var[missing_more_than_name] + \
                                   by_id_var[missing_tolerance_count_name]

    by_id_var[missing_more_than_pct_str(missing_tolerance, missing_quantile, missing_display_str, id_var)] = \
        (by_id_var[missing_more_than_name] /
        by_id_var[id_count_name]) * 100

    by_id_var.drop([
        missing_more_than_name,
        missing_tolerance_count_name
    ], axis=1, inplace=True)

    return by_id_var


def _by_id_count_long_df(df: pd.DataFrame, byvars: StrList, id_var: str,
                         count_with_missings_var: str, missing_tolerance: IntOrNone = 0,
                         missing_quantile: IntOrNone = None,
                         missing_display_str: str = 'Missing') -> pd.DataFrame:
    """
    Note: uses missing_quantile if specified, else uses missing_tolerance

    Args:
        df:
        byvars:
        id_var:
        count_with_missings_var:
        missing_tolerance:
        missing_quantile:
        missing_display_str:

    Returns:

    Raises:
        ValueError: if missing_tolerance and missing_quantile are both None
        ValueError: if columns _one or _pct are in df

    """

    if '_one' in df.columns or '_pct' in df.columns:
        raise ValueError('will override column _one and _pct in df')

    df['_one'] = 1  # temporary variable for counting without missing
    by_firm_counts = df.groupby([id_var] + byvars)[['_one', count_with_missings_var]].count().reset_index()
    df.drop('_one', axis=1, inplace=True)

    if missing_quantile is not None:
        by_firm_counts['_pct'] = by_firm_counts[count_with_missings_var] / by_firm_counts['_one']  # percentage coverage
        minimum_allowed_coverage = (1 - missing_quantile)
        missing_df = by_firm_counts[by_firm_counts['_pct'] < minimum_allowed_coverage]
        full_df = by_firm_counts[by_firm_counts['_pct'] >= minimum_allowed_coverage]
    elif missing_tolerance is not None:
        missing_df = by_firm_counts[by_firm_counts[count_with_missings_var] + missing_tolerance < by_firm_counts['_one']]
        full_df = by_firm_counts[by_firm_counts[count_with_missings_var] + missing_tolerance >= by_firm_counts['_one']]
    else:
        raise ValueError('pass one of missing_tolerance and missing_quantile. got both as None')

    name_args = (missing_tolerance, missing_quantile, missing_display_str, id_var)

    missing_counts = missing_df.groupby(byvars)[id_var].count()
    missing_counts.name = missing_more_than_str(*name_args)

    full_counts = full_df.groupby(byvars)[id_var].count()
    full_counts.name = missing_tolerance_count_str(*name_args)

    by_id_var = pd.concat([missing_counts, full_counts], axis=1).fillna(0)
    return by_id_var.reset_index()


import pandas as pd
from typing import Callable
from functools import partial

from datacode.display import display_df_dict
from datacode.typing import DfDictOrNone, FloatList


def format_numbers_to_decimal_places(item, decimals=2, coerce_ints: bool = False):

    if isinstance(item, (float, int)):
        if abs(item) > 999999.99:
            # Millions formatter
            item = item / 1000000
            # For millions decimals, treat int as float (still have decimals, because 7.02M is better than 7M for int
            return f'{item:,.{decimals}f}M'
        if coerce_ints:
            decimals = 0 if int(item) == item else decimals  # checks if is int stored as type float
        else:
            decimals = decimals if isinstance(item, float) else 0  # uses dtype to handle int vs. float
        return f'{item:,.{decimals}f}'
    else:
        return item

def describe_df(df: pd.DataFrame, disp: bool=True, format_func: Callable=format_numbers_to_decimal_places,
                format_kwargs: dict=None, percentiles: FloatList =None) -> DfDictOrNone:
    """

    Args:
        df:
        disp: True to display summaries, False to only return dict of summaries

    Returns:

    """
    if format_kwargs is None:
        format_kwargs = {}

    if percentiles is None:
        percentiles = [0.01, 0.05, 0.1, 0.25, .75, .9, .95, 0.99]

    summaries_dict = {}
    for dtype, count in df.dtypes.value_counts().iteritems():
        dtype_str = str(dtype)
        summary: pd.DataFrame = df.describe(include=dtype_str, percentiles=percentiles).T
        if 'int' in dtype_str:
            summary = summary.astype(int)  # describe is always float for int dtype, cast output format back to int
        summary_name = f'{dtype_str} Variables'
        summaries_dict[summary_name] = summary

    full_format_func = partial(format_func, **format_kwargs)

    summaries_dict = {
        summary_name: summary.applymap(full_format_func) for summary_name, summary in summaries_dict.items()
    }

    if disp:
        display_df_dict(summaries_dict)
    else:
        return summaries_dict






import math
from typing import Union, Tuple
from datacode.formatters.stars import parse_stars_value


class Result:

    def __init__(self, value: str):
        self.orig_result = value
        self.str_result, self.result, self.num_stars = parse_value(value)

    def __repr__(self):
        return f'<Result({self})>'

    def __str__(self):
        return f'{self.str_result}{"*" * self.num_stars}'

    @property
    def is_significant(self) -> bool:
        return self.num_stars > 0

    def is_significant_at(self, pct: int) -> bool:
        """

        Args:
            pct: 90 - 99 of significance level, e.g. 90, 95, or 99

        Returns: bool

        """
        if pct < 90:
            raise ValueError('cannot determine significance below 90%')
        elif pct < 95:
            return self.num_stars > 0
        elif pct < 99:
            return self. num_stars > 1
        elif pct == 99:
            return self. num_stars > 2
        else:
            raise ValueError('cannot determine significance above 99%')

    @property
    def is_positive(self) -> bool:
        try:
            return self.result > 0
        except TypeError:
            return False

    @property
    def is_negative(self) -> bool:
        try:
            return self.result < 0
        except TypeError:
            return False

    def as_logit_result(self) -> 'LogitResult':
        lr = LogitResult(self.orig_result)
        return lr


class LogitResult(Result):

    @property
    def odds_ratio(self) -> float:
        return math.exp(self.result)

    @property
    def probability_relative_change(self) -> float:
        return self.odds_ratio - 1

    @property
    def probability_relative_change_str(self) -> str:
        rel_change = self.probability_relative_change
        if rel_change > 0:
            change_str = 'increase'
        else:
            change_str = 'decrease'
        rel_change = abs(rel_change) * 100
        return f'{rel_change:.1f}% {change_str}'


def parse_value(value: str) -> Tuple[str, Union[str, float], int]:
    result, stars = parse_stars_value(value)
    num_result = to_numeric_if_numeric(result)
    num_stars = len(stars)
    return result, num_result, num_stars



def to_numeric_if_numeric(value: str) -> Union[str, float]:
    try:
        return float(value)
    except ValueError:
        return value


import re
from typing import Tuple, TYPE_CHECKING, List

import pandas as pd
from sympy import Symbol

from datacode.models.transform.transform import Transform
from pd_utils.filldata import add_missing_group_rows

from datacode.models.dtypes.datetime_type import DatetimeType

if TYPE_CHECKING:
    from datacode.models.variables.variable import Variable
    from datacode.models.column.column import Column
    from datacode.models.source import DataSource

LAG_NAME_PATTERN = re.compile(r'(.+)\$_{t - (\d+)}\$')
LAG_SYMBOL_PATTERN = re.compile(r'(.+)_{t - (\d+)}')


def varname_to_lagged_varname(varname: str, num_lags: int = 1) -> str:
    if num_lags == 0:
        # No lag string necessary
        return varname

    try:
        base_var, existing_lags = lag_varname_to_varname_and_lag(varname)
    except VariableIsNotLaggedVariableException:
        # Variable is not already lagged, so just add lag portion to str
        return _varname_to_lagged_varname(varname, num_lags)

    # Variable was lagged originally, need to add an additional number of lags and apply to base name
    total_lags = existing_lags + num_lags
    return _varname_to_lagged_varname(base_var, total_lags)


def _varname_to_lagged_varname(varname: str, num_lags: int = 1) -> str:
    return varname + f"$_{{t - {num_lags}}}$"


def lag_varname_to_varname_and_lag(varname: str) -> Tuple[str, int]:
    match = LAG_NAME_PATTERN.match(varname)
    if match is None:
        raise VariableIsNotLaggedVariableException(f'could not parse {varname} as a lagged name')

    base_varname = match.group(1)
    lag_num = int(match.group(2))
    return base_varname, lag_num


def var_symbol_to_lagged_var_symbol(varname: str, num_lags: int = 1) -> str:
    if num_lags == 0:
        # No lag string necessary
        return varname

    try:
        base_var, existing_lags = lag_var_symbol_to_var_symbol_and_lag(varname)
    except VariableIsNotLaggedVariableException:
        # Variable is not already lagged, so just add lag portion to str
        return _var_symbol_to_lagged_var_symbol(varname, num_lags)

    # Variable was lagged originally, need to add an additional number of lags and apply to base name
    total_lags = existing_lags + num_lags
    return _var_symbol_to_lagged_var_symbol(base_var, total_lags)


def _var_symbol_to_lagged_var_symbol(varname: str, num_lags: int = 1) -> str:
    return varname + f"_{{t - {num_lags}}}"


def lag_var_symbol_to_var_symbol_and_lag(varname: str) -> Tuple[str, int]:
    match = LAG_SYMBOL_PATTERN.match(varname)
    if match is None:
        raise VariableIsNotLaggedVariableException(f'could not parse {varname} as a lagged name')

    base_varname = match.group(1)
    lag_num = int(match.group(2))
    return base_varname, lag_num


def _create_lagged_variable_panel(df: pd.DataFrame, col: str, by_vars: List[str], num_lags: int = 1):
    """
    Note: inplace
    """
    df[col] = df.groupby(by_vars)[col].shift(num_lags)


def _create_lagged_variable(df: pd.DataFrame, col: str, num_lags: int = 1) -> None:
    """
    Note: inplace
    """
    df[col] = df[col].shift(num_lags)


class VariableIsNotLaggedVariableException(Exception):
    pass


class DidNotFindSingleTimeIndexException(Exception):
    pass


def create_lags_transform_name_func(name: str, num_lags: int = 1, **kwargs) -> str:
    return varname_to_lagged_varname(name, num_lags=num_lags)


def create_lags_transform_symbol_func(sym: Symbol, num_lags: int = 1, **kwargs) -> str:
    sym_str = str(sym)
    new_sym_str = var_symbol_to_lagged_var_symbol(sym_str, num_lags=num_lags)
    return Symbol(new_sym_str)


def create_lags_transform_data_func(col: 'Column', variable: 'Variable', source: 'DataSource',
                                    num_lags: int = 1, fill_missing_rows: bool = True,
                                    fill_method="ffill", fill_limit: int = None) -> 'DataSource':
    orig_cols = list(source.df.columns)  # copy column order for reordering at end

    can_fill_missing_rows = False
    by_vars = []
    time_vars = []
    if col.indices:
        # Try to get time index
        time_indices = [
            col_idx for col_idx in col.indices if all(var.dtype == DatetimeType() for var in col_idx.variables)
        ]
        if len(time_indices) > 1:
            raise DidNotFindSingleTimeIndexException(f'found time indices {time_indices}, not clear how to lag')
        elif len(time_indices) != 0:
            # Got a time index, so can fill missing rows in that index,
            # so long as there is an id index (determined next)
            can_fill_missing_rows = True
        for col_idx in time_indices:
            time_vars.extend(col_idx.variables)

        # Get other indices
        other_indices = [col_idx for col_idx in col.indices if col_idx not in time_indices]
        if len(other_indices) > 0:
            # Got other indices
            for col_idx in other_indices:
                by_vars.extend(col_idx.variables)
        else:
            # Not by any index, so no reason to fill missing rows
            can_fill_missing_rows = False

    by_var_names = [var.name for var in by_vars]
    time_var_names = [var.name for var in time_vars]
    lag_kwargs = dict(num_lags=num_lags)
    if by_vars:
        lag_kwargs.update(by_vars=by_var_names)
        lag_func = _create_lagged_variable_panel
        source.df.sort_values(by_var_names + time_var_names, inplace=True)

        if fill_missing_rows and can_fill_missing_rows:
            # Save original index, for outputting df of same shape
            orig_index_df = source.df[[]]

            # TODO [#61]: use implementation of add_missing_group_rows which does not require dropping and resetting index
            #
            # Need to wait for pd_utils to support it
            orig_index_names = orig_index_df.index.names
            source.df = add_missing_group_rows(
                source.df.reset_index(), by_var_names, time_var_names, fill_method=fill_method, fill_limit=fill_limit
            ).set_index(orig_index_names)
    else:
        lag_func = _create_lagged_variable  # type: ignore

    lag_func(source.df, variable.name, **lag_kwargs)

    if by_vars and fill_missing_rows and can_fill_missing_rows:
        # Don't want to expand size of df
        # For some reason, join doesn't work correctly. So reset indexes, merge, and set indexes back
        source.df = orig_index_df.reset_index().merge(
            source.df.reset_index(), how='left', on=orig_index_df.index.names
        ).set_index(orig_index_df.index.names)
        # Reorder back to original order of columns
        source.df = source.df[orig_cols]

    return source


lag_transform = Transform(
    'lag',
    name_func=create_lags_transform_name_func,
    data_func=create_lags_transform_data_func,
    symbol_func=create_lags_transform_symbol_func,
    data_func_target='source'
)


from datacode.summarize.subset.outliers.typing import DfDict
from datacode.typing import DocumentOrTable
import pyexlatex.table as lt
from pyexlatex import Document

def df_dict_to_table(df_dict: DfDict, as_document: bool=False,
                     author: str=None, caption: str=None) -> DocumentOrTable:

    table = lt.Table.from_panel_name_df_dict(
        df_dict,
        include_index=True,
        top_left_corner_labels='Variable',
        caption=caption,
        landscape=True
    )

    if as_document:
        return Document(
            table,
            author=author,
            title=caption
        )

    return table


from itertools import product

from .models.interface import IDSet, IDCollection, IDComparison, IDComparisonCollection


def compare_id_collections(id_collection_1: IDCollection, id_collection_2: IDCollection, name=None) -> IDComparisonCollection:
    comparisons = []
    for id_set1, id_set2 in product(id_collection_1, id_collection_2):
        if id_set1.name == id_set2.name:
            comparisons.append(compare_id_sets(id_set1, id_set2, name=id_set1.name))
        # If not same name, not same type of id so continue
    return IDComparisonCollection(comparisons, name=name)

def compare_id_sets(id_set1: IDSet, id_set2: IDSet, name=None):
    compare_values = _compare_id_sets(id_set1, id_set2)
    return IDComparison(*compare_values, name=name)

def _compare_id_sets(id_set1: IDSet, id_set2: IDSet):
    differences_intersection = _set_differences_and_intersection(id_set1, id_set2)
    return [len(set_) for set_ in differences_intersection]

def _set_differences_and_intersection(id_set1: IDSet, id_set2: IDSet):
    """

    :param id_set1:
    :type id_set1:
    :param id_set2:
    :type id_set2:
    :return: differences, tuple of three sets. First is elements of 1 not in 2, second is elements in both, third is
        elements in 2 that are not in 1
    :rtype: tuple
    """
    return id_set1.difference(id_set2), id_set1.intersection(id_set2), id_set2.difference(id_set1)



import pandas as pd
from datacode.typing import StrList


def obs_pct_long_df(df: pd.DataFrame, byvars: StrList, obs_count_var: str,
                    count_with_missings_var: str, missing_display_str: str='Missing') -> pd.DataFrame:
    """

    Args:
        df:
        byvars:
        obs_count_var: should never be missing. may be id var, or may have to create a column
        count_with_missings_var:

    Returns:

    """

    long_obs_df = _obs_count_long_df(
        df,
        byvars,
        [obs_count_var, count_with_missings_var]
    )

    long_obs_df[f'Obs {missing_display_str} Percentage'] = \
        (1 - (long_obs_df[count_with_missings_var] / long_obs_df[obs_count_var])) * 100
    long_obs_df.drop(count_with_missings_var, axis=1, inplace=True)
    long_obs_df.rename(columns={obs_count_var: 'Obs'}, inplace=True)

    return long_obs_df.reset_index()

def _obs_count_long_df(df: pd.DataFrame, byvars: StrList, countvars: StrList) -> pd.DataFrame:
    return df.groupby(byvars)[countvars].count()


from functools import partial
from typing import Callable, Any, Dict, Optional, Sequence

from datacode.models.pipeline.operations.merge import MergeOptions
from datacode.models.pipeline.base import DataPipeline
from datacode.models.types import DataSourcesOrPipelines


class DataMergePipeline(DataPipeline):
    """
    Handles data pipelines involving merges between two or more sources or pipelines
    """

    def __init__(self, data_sources: DataSourcesOrPipelines = None,
                 merge_options_list: Optional[Sequence[MergeOptions]] = None,
                 post_merge_cleanup_func: Optional[Callable] = None,
                 name: Optional[str] = None, cleanup_kwargs: Optional[Dict[str, Any]] = None):

        if cleanup_kwargs is None:
            cleanup_kwargs = {}
        if merge_options_list is None:
            merge_options_list = []
        if data_sources is None:
            data_sources = []

        self._set_cleanup_func(post_merge_cleanup_func, **cleanup_kwargs)
        self.cleanup_kwargs = cleanup_kwargs
        self.data_sources = data_sources

        super().__init__(data_sources, merge_options_list, name=name)

    def _validate(self):
        self._validate_data_sources_merge_options()

    def _validate_data_sources_merge_options(self):
        if len(self.data_sources) - 1 != len(self.operation_options):
            raise ValueError(f'must have one fewer merge options than data sources. Have {len(self.data_sources)} data '
                             f'sources and {len(self.operation_options)} merge options.')

    def execute(self):
        self._validate()

        super().execute(output=False)

        if self.has_post_merge_cleanup_func:
            self.df = self.post_merge_cleanup_func(self.df)
        self.output()

        return self.df

    def _set_cleanup_func(self, post_merge_cleanup_func, **cleanup_kwargs):
        if post_merge_cleanup_func is not None:
            self.has_post_merge_cleanup_func = True
            self.post_merge_cleanup_func = partial(post_merge_cleanup_func, **cleanup_kwargs)
        else:
            self.has_post_merge_cleanup_func = False




from typing import Callable, TYPE_CHECKING
if TYPE_CHECKING:
    from datacode.models.pipeline.operations.merge import DataMerge

from datacode.models.logic.merge.summarize import get_summary_of_df
from datacode.display import display_df_dict


def display_merge_summary(merge: 'DataMerge', *summary_args, summary_method: str=None, summary_function: Callable=None,
                          summary_attr: str=None, **summary_method_kwargs):
    # If nothing is passed to use for summary, use df.head()
    if (summary_attr is None) and (summary_function is None) and (summary_method is None):
        summary_method = 'head'

    df_disp_dict = _disp_df_dict_from_merge(
        merge,
        *summary_args,
        summary_method=summary_method,
        summary_function=summary_function,
        summary_attr=summary_attr,
        **summary_method_kwargs
    )

    if summary_attr is not None:
        summary_disp = f'df.{summary_attr}'
    if summary_function is not None:
        summary_disp = f'{summary_function.__name__}(df, *{summary_args}, **{summary_method_kwargs})'
    if summary_method is not None:
        summary_disp = f'df.{summary_method}(*{summary_args}, **{summary_method_kwargs})'

    display_df_dict({
        f'{summary_disp} called on: ' + merge.merge_str: df_disp_dict
    })


def _disp_df_dict_from_merge(merge, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):

    # keys are names of dataframes, values are dataframes themselves
    df_dict = _df_dict_from_merge(merge)

    return {
        name: get_summary_of_df(
            df,
            *summary_args,
            summary_method=summary_method,
            summary_function=summary_function,
            summary_attr=summary_attr,
            **summary_method_kwargs
        ) for name, df in df_dict.items()
    }


def _df_dict_from_merge(merge: 'DataMerge'):
    df_dict = {
        f'Left Dataset: {merge.data_sources[0].name}': merge.data_sources[0].df,
        f'Right Dataset: {merge.data_sources[1].name}': merge.data_sources[1].df,
        f'Result: {merge.result.name}': merge.result.df
    }

    return df_dict


import pandas as pd
from typing import List

from pd_utils.pdutils import _to_list_if_str
from datacode.ff.fftypes import (
    DictofStrsandStrLists,
    TwoStrTuple,
    ListOrStr,
    StrBoolDict,
    StrOrInt
)
from datacode.ff.create.model import parse_model

def construct_minus_variables(df: pd.DataFrame, labels: DictofStrsandStrLists, pairing: TwoStrTuple,
                              factor_model: StrOrInt=3, byvars: ListOrStr = None, datevar='Date',
                              size_var: str=None, value_var: str=None, profitability_var: str=None,
                              investment_var: str=None,
                              custom_low_minus_high_dict: StrBoolDict=None) -> pd.DataFrame:

    byvars = _to_list_if_str(byvars)

    low_minus_high_dict = _get_low_minus_high_dict(
        factor_model=factor_model,
        size_var=size_var,
        value_var=value_var,
        profitability_var=profitability_var,
        investment_var=investment_var,
        custom_low_minus_high_dict=custom_low_minus_high_dict
    )

    minus_vars = []
    for index, portvar in enumerate(pairing):
        minus_vars.append(
            _construct_minus_variable(
                df,
                labels=labels,
                portvar_index=index,
                portvar=portvar,
                byvars=byvars,
                datevar=datevar,
                low_minus_high=low_minus_high_dict[portvar]
            )
        )

    # Rename first factor to show it was calculated with second factor. E.g. SMB -> SMB_HML
    new_main_portname = _calculated_with_varname(minus_vars[0], minus_vars[1])
    df.rename(columns={minus_vars[0]: new_main_portname}, inplace=True)
    minus_vars = [new_main_portname] + minus_vars[1:]

    if byvars is not None:
        all_vars = byvars + [datevar] + minus_vars
    else:
        all_vars = [datevar] + minus_vars

    return df[all_vars]


def _get_low_minus_high_dict(factor_model: StrOrInt=3, size_var: str=None, value_var: str=None,
               profitability_var: str=None, investment_var: str=None,
               custom_low_minus_high_dict: StrBoolDict=None) -> StrBoolDict:

    factor_model = parse_model(factor_model)

    if factor_model == 'custom':
        return custom_low_minus_high_dict

    low_minus_high_dict = {
        size_var: True,
        value_var: False
    }

    if profitability_var is not None:
        low_minus_high_dict.update({
            profitability_var: False,
            investment_var: True
        })

    return low_minus_high_dict

def _construct_minus_variable(df: pd.DataFrame, labels: DictofStrsandStrLists, portvar_index=0,
                              portvar: str='Market Equity', byvars: ListOrStr=None, datevar='Date',
                              low_minus_high=False) -> str:

    long_label, short_label = _set_long_label_short_label(labels, portvar=portvar, low_minus_high=low_minus_high)
    long_cols = _get_port_cols_matching_label_at_index(
        df, port_label=long_label, port_index=portvar_index, byvars=byvars, datevar=datevar
    )
    short_cols = _get_port_cols_matching_label_at_index(
        df, port_label=short_label, port_index=portvar_index, byvars=byvars, datevar=datevar
    )

    minus_label = _minus_label(long_label, short_label)
    df[minus_label] = df[long_cols].apply(lambda x: x.mean(), axis=1) - df[short_cols].apply(lambda x: x.mean(), axis=1)

    return minus_label


def _set_long_label_short_label(labels: DictofStrsandStrLists, portvar: str, low_minus_high: bool=False) -> TwoStrTuple:
    portvar_labels = labels[portvar]

    if low_minus_high:
        # long bottom portfolio, short top portfolio
        long_label = portvar_labels[0]
        short_label = portvar_labels[-1]
    else:
        # short bottom portfolio, long top portfolio
        long_label = portvar_labels[-1]
        short_label = portvar_labels[0]

    return long_label, short_label

def _get_port_cols_matching_label_at_index(df: pd.DataFrame, port_label: str='S', port_index=0,
                                  byvars: ListOrStr=None, datevar='Date') -> List[str]:

    if byvars is None:
        byvars = []

    port_cols = [col for col in df.columns if col not in byvars + [datevar]]
    return [col for col in port_cols if col[port_index] == port_label]

def _calculated_with_varname(orig_var: str, calculated_with: str):
    return f'{orig_var}_{calculated_with}'

def _calculated_with_varname_to_orig_var(calc_with_varname: str) -> str:
    return calc_with_varname.split('_')[0]

def _get_minus_label(labels: DictofStrsandStrLists,
                 portvar: str='Market Equity', low_minus_high=False) -> str:
    long_label, short_label = _set_long_label_short_label(labels, portvar=portvar, low_minus_high=low_minus_high)
    return _minus_label(long_label, short_label)

def _minus_label(long_label: str, short_label: str) -> str:
    return f'{long_label}M{short_label}'


import pandas as pd

from typing import List
from datacode.ff.fftypes import StrOrInt, DictofStrsandStrLists
from datacode.ff.create.model import parse_model

def get_and_set_labels(df: pd.DataFrame, factor_model: StrOrInt=3, size_var: str=None, value_var: str=None,
               profitability_var: str=None, investment_var: str=None,
               custom_labels: DictofStrsandStrLists=None) -> DictofStrsandStrLists:

    labels = _get_labels(
        factor_model=factor_model,
        size_var=size_var,
        value_var=value_var,
        profitability_var=profitability_var,
        investment_var=investment_var,
        custom_labels=custom_labels
    )

    _set_labels(df, labels)

    return labels

def _set_labels(df: pd.DataFrame, labels: DictofStrsandStrLists):
    for col, labels in labels.items():
        _set_label(df, col, labels)

def _set_label(df: pd.DataFrame, label_var: str, labels: List[str]) -> None:
    """
    Note: inplace
    """
    assert len(labels) == len(df[label_var].unique())

    for i, label_value in enumerate(labels):
        port_num = i + 1 # portfolios are 1-indexed, not 0-indexed. portfolio 1 is lowest value portfolio
        df.loc[df[label_var] == port_num, label_var] = label_value

def _get_labels(factor_model: StrOrInt=3, size_var: str=None, value_var: str=None,
               profitability_var: str=None, investment_var: str=None,
               custom_labels: DictofStrsandStrLists=None) -> DictofStrsandStrLists:
    """
    Note: inplace
    """
    factor_model = parse_model(factor_model)

    if factor_model == 'custom':
        labels = custom_labels
    else:
        labels = _get_default_labels(
            size_var=size_var,
            value_var=value_var,
            profitability_var=profitability_var,
            investment_var=investment_var
        )

    return labels


def _get_default_labels(size_var: str, value_var: str,
                        profitability_var: str=None, investment_var: str=None) -> DictofStrsandStrLists:
    default_labels = {
        size_var: ['S','B'],
        value_var: ['L','M','H']
    }

    if profitability_var is not None:
        default_labels.update({
            profitability_var: ['W','M','R'],
        })

    if investment_var is not None:
        default_labels.update({
            investment_var: ['C','M','A'],
        })

    return default_labels


from typing import Union, Sequence, Dict, Any, List

import pandas as pd

from datacode.models.variables import Variable
from datacode.models.source import DataSource
from datacode.sem.df import assemble_model_data
from datacode.sem.latents import get_latents
from datacode.sem.model import model_str
from datacode.sem.run import run_model
from datacode.sem.summary import SEMSummary


class SEM:

    def __init__(self, data: Union[DataSource, pd.DataFrame], structural_dict: Dict[Variable, Sequence[Variable]],
                 measurement_dict: Dict[Variable, Sequence[Variable]],
                 var_corr_groups: Sequence[Sequence[Variable]], scale: bool = True,
                 robust_scale: bool = False, pos_def_warnings: str = 'ignore', **scale_kwargs):
        self.data = data
        self.structural_dict = structural_dict
        self.measurement_dict = measurement_dict
        self.var_corr_groups = var_corr_groups
        self.scale = scale
        self.robust_scale = robust_scale
        self.scale_kwargs = scale_kwargs
        self.pos_def_warnings = pos_def_warnings
        self.model_def = model_str(structural_dict, measurement_dict, var_corr_groups)

        self.opt = None
        self.model = None
        self.summary = None
        self._model_df = None

    def fit(self, **opt_kwargs) -> SEMSummary:
        self.opt = run_model(self.model_def, self.model_df, pos_def_warnings=self.pos_def_warnings, **opt_kwargs)
        self.model = self.opt.model
        self.summary = SEMSummary(
            self.opt,
            self.observed_endog_vars,
            self.structural_dict,
            self.measurement_dict,
            self.var_corr_groups,
            self.all_vars,
            self.scale,
            self.robust_scale
        )
        return self.summary

    @property
    def df(self) -> pd.DataFrame:
        if isinstance(self.data, pd.DataFrame):
            return self.data
        elif isinstance(self.data, DataSource):
            return self.data.df
        else:
            raise ValueError(f'did not pass DataSource or DataFrame for data, got type {type(self.data)}')

    @property
    def model_df(self) -> pd.DataFrame:
        if self._model_df is None:
            self._model_df = assemble_model_data(
                self.df,
                self.observed_vars,
                scale=self.scale,
                robust_scale=self.robust_scale,
                **self.scale_kwargs
            )
        return self._model_df

    @property
    def observed_endog_vars(self) -> Sequence[Variable]:
        return [var for var in self.structural_dict.keys() if var not in self.measurement_dict.keys()]

    @property
    def latent_vars(self) -> Sequence[Variable]:
        return list(self.measurement_dict.keys())

    @property
    def all_vars(self) -> List[Variable]:
        collected = []
        for var_dict in [self.structural_dict, self.measurement_dict]:
            for lhs_var, rhs_vars in var_dict.items():
                collected.append(lhs_var)
                collected.extend(rhs_vars)
        for var_seq in self.var_corr_groups:
            collected.extend(var_seq)
        return list(set(collected))

    @property
    def observed_vars(self) -> List[Variable]:
        return [var for var in self.all_vars if var not in self.latent_vars]

    @property
    def latents(self) -> pd.DataFrame:
        return get_latents(self.opt, self.model_df, self.latent_vars)




from typing import Sequence, Dict, Optional, Type

from datacode.models.dtypes.base import DataType


class DataTypeManager:

    def __init__(self, data_types: Sequence[Type[DataType]]):
        self.data_types = data_types
        self.name_map = self._create_name_map()
        self.root_name_map = self._create_name_map('name_roots', existing_maps=[self.name_map])

    def _create_name_map(self, name_attr: str = 'names',
                         existing_maps: Optional[Sequence[Dict[str, Type[DataType]]]] = None
                         ) -> Dict[str, Type[DataType]]:
        all_maps = {}
        if existing_maps is not None:
            for e_map in existing_maps:
                all_maps.update(e_map)

        name_map = {}
        for data_type in self.data_types:
            for name in getattr(data_type, name_attr):
                if name in all_maps:
                    raise ValueError(f'Got name {name} for data type {data_type} when it was already '
                                     f'used by data type {name_map[name]}')
                name_map[name] = data_type
                all_maps[name] = data_type
        return name_map

    def get_by_name(self, name: str) -> DataType:
        name = name.lower()
        # Try for exact match first
        try:
            cls = self.name_map[name]
            return cls.from_str(name)
        except KeyError:
            pass

        # Now look up in root name dict
        for dt_name, dt in self.root_name_map.items():
            if name.startswith(dt_name):
                cls = dt
                return cls.from_str(name)

        # Didn't find in either dict
        raise KeyError(name)



from typing import Dict, TYPE_CHECKING, Optional, Callable, Any

from sympy import Symbol

if TYPE_CHECKING:
    from datacode.models.variables.collection import VariableCollection, Variable

PrefixVariableCollectionDict = Dict[str, 'VariableCollection']
StrFunc = Optional[Callable[[str], str]]
ValueFunc = Optional[Callable[[Any], Any]]
SymbolFunc = Optional[Callable[[Symbol], Symbol]]



import datetime

from pandas.testing import assert_frame_equal

from datacode import DataSource
from tests.pipeline.base import PipelineTest, EXPECT_GENERATED_DF


class TestDataGeneratorPipeline(PipelineTest):

    def test_create_and_run_generator_pipeline_from_func(self):
        dgp = self.create_generator_pipeline()
        dgp.execute()

        assert_frame_equal(dgp.df, EXPECT_GENERATED_DF)

    def test_auto_run_pipeline_by_load_source_with_no_location(self):
        dgp = self.create_generator_pipeline()

        ds = DataSource(pipeline=dgp, location=self.csv_path_output)
        df = ds.df
        assert_frame_equal(df, EXPECT_GENERATED_DF)

    def test_auto_run_pipeline_by_load_source_with_newer_pipeline(self):
        now = datetime.datetime.now()
        later = now + datetime.timedelta(minutes=5)
        dgp = self.create_generator_pipeline(last_modified=later)

        self.create_csv()
        ds = self.create_source(pipeline=dgp, df=None)
        df = ds.df
        assert dgp._operation_index == 1
        assert_frame_equal(df, EXPECT_GENERATED_DF)



import pandas as pd
from datacode.psm.typing import DfDict, StrList, StrOrNone, StrListOrNone
import pyexlatex.table as lt
from datacode.psm.predict import explain_probability_of_treatment
from datacode.psm.summarize.latex import matching_latex_table_from_df_dict
from datacode.psm.summarize.predictions import summarize_predictions
from datacode.psm.summarize.stats import matching_summary_stats
from datacode.psm.names import get_prob_treated_varname
from datacode.psm.typing import FloatOrNone

def create_and_output_matching_summary_latex_table(df: pd.DataFrame, matched_df: pd.DataFrame, predict_df: pd.DataFrame,
                                                   treated_var: str, xvars: StrList, entity_var: str, time_var: str,
                                                   fe: StrListOrNone = None,
                                                   prob_treated_var: StrOrNone = None, control_name: str = 'Control',
                                                   treated_name: str = 'Treated', below_text: StrOrNone = None,
                                                   outfolder: str = '.', caption: str = 'Propensity Score Matching',
                                                   min_matching_pct: FloatOrNone = None,
                                                   ) -> lt.Table:
    summary_df_dict = create_matching_summary_df_dict(
        df,
        matched_df,
        predict_df,
        treated_var,
        xvars,
        entity_var,
        fe=fe,
        prob_treated_var=prob_treated_var,
        control_name=control_name,
        treated_name=treated_name
    )

    table = matching_latex_table_from_df_dict(
        summary_df_dict,
        entity_var,
        time_var,
        below_text=below_text,
        caption=caption,
        min_matching_pct=min_matching_pct
    )

    table.to_pdf_and_move(
        outfolder=outfolder,
        outname=caption
    )

    return table


def create_matching_summary_df_dict(df: pd.DataFrame, matched_df: pd.DataFrame, predict_df: pd.DataFrame,
                                    treated_var: str, xvars: StrList, entity_var: str, fe: StrListOrNone = None,
                                    prob_treated_var: StrOrNone = None, control_name: str = 'Control',
                                    treated_name: str = 'Treated') -> DfDict:
    prob_treated_var = get_prob_treated_varname(treated_var, prob_treated_var_name=prob_treated_var)
    _, model_summ = explain_probability_of_treatment(
        df,
        treated_var,
        xvars,
        fe=fe
    )

    predict_summ = summarize_predictions(
        predict_df,
        treated_var,
        treated_name=treated_name,
        control_name=control_name
    )

    describe_vars = [treated_var, prob_treated_var] + xvars
    match_summ = matching_summary_stats(
        df,
        matched_df,
        treated_var,
        describe_vars,
        entity_var,
        control_name=control_name,
        treated_name=treated_name
    )

    return {
        'model': model_summ.tables[0],
        'predict': predict_summ,
        'match': match_summ
    }


import pyexlatex.table as lt
from datacode.typing import SimpleDfDict, StrOrNone
from datacode.summarize.subset.missing.detail.textfuncs import num_or_pct_word, pct_of_if_necessary
from datacode.typing import SimpleDfDict, StrOrNone, IntSequenceOrNone, IntOrNone, FloatSequenceOrNone, FloatOrNone

def missing_detail_df_dict_to_table_and_output(df_dict: SimpleDfDict, summary_panel: lt.Panel,
                                               row_byvar, col_byvar: str,
                                               id_var: str,
                                               count_with_missings_var: str,
                                               missing_tolerance: IntOrNone = 0, missing_quantile: FloatOrNone = None,
                                               summary_missing_tolerances: IntSequenceOrNone = (0, 5, 10),
                                               summary_missing_quantiles: FloatSequenceOrNone = None,
                                               missing_display_str: str = 'Missing',
                                               period_display_name: str='Period',
                                               extra_caption: str='', extra_below_text: str='',
                                               align: str=None,
                                               outfolder: StrOrNone=None) -> lt.Table:

    caption = f'Data Gap Analysis - {missing_display_str} {count_with_missings_var}'

    missing_display_str = missing_display_str.lower()

    if missing_display_str == 'missing':
        missing_long_display_str = 'missing information'
    else:
        missing_long_display_str = missing_display_str

    # percentage or number string
    summary_num_or_pct = num_or_pct_word(summary_missing_tolerances, summary_missing_quantiles)
    # % of if quantiles, empty string if tolerances
    percent_of_or_blank = pct_of_if_necessary(missing_tolerance, missing_quantile)

    below_text = f"""
    This table shows where the {count_with_missings_var} variable is {missing_long_display_str}. 
    For Panels A, B, C, and D, each item represents a subsample analysis where {row_byvar} is the value given by the row
    and where {col_byvar} is the value given by the column, while Panel E is a full sample analysis.
    Panel A describes the number of observations.
    Panel B describes the percentage of observations with {missing_long_display_str} for {count_with_missings_var}.
    Panel C describes the number of unique {id_var}s. 
    Panel D describes the percentage of unique {id_var}s which have more than {missing_tolerance}{percent_of_or_blank} 
    observations with {missing_long_display_str} for {count_with_missings_var}. 
    Panel E has four subsections. The first section shows the count of observations and the percentage that are 
    {missing_long_display_str}. The second section shows the count of unique {id_var}s. The third section shows
    summary statistics for the number of {period_display_name}s per {id_var}. The fourth section shows the percentage
    of {id_var}s which have at least the {summary_num_or_pct} of {period_display_name}s given by the column. 
    """ + extra_below_text

    if extra_caption:
        caption = f'{caption} - {extra_caption}'

    detail_data_tables = {
        name: lt.DataTable.from_df(df, include_index=True, extra_header=col_byvar)
        for name, df in df_dict.items()
    }
    detail_panels = [lt.Panel.from_data_tables([dt], name=name) for name, dt in detail_data_tables.items()]

    table = lt.Table.from_panel_list(
        detail_panels + [summary_panel],
        label_consolidation='str',
        below_text=below_text,
        caption=caption,
        align=align,
        top_left_corner_labels = row_byvar
    )

    if outfolder is not None:
        table.to_pdf_and_move(
            outname=caption,
            outfolder=outfolder
        )

    return table


from typing import Optional
import pandas as pd
import numpy as np


def collect_portfolios_through_time(df: pd.DataFrame, portvar: str, id_var: str, needed_calendar_days: int,
                                    needed_prior_calendar_days: Optional[int] = None,
                                    datevar: str = 'Date', portfolio_datevar: str = 'Portfolio Date'
                                    ) -> pd.DataFrame:
    output_df = pd.DataFrame()
    # TODO [#4]: collect portfolios through time make more efficient
    for port_date in df[portfolio_datevar].unique():
        port_assignments = df.loc[
            df[portfolio_datevar] == port_date,
            [id_var, portvar]
        ].drop_duplicates()
        extended_date = port_date + np.timedelta64(needed_calendar_days, 'D')
        begin_date = _portfolio_begin_date(port_date, needed_prior_calendar_days=needed_prior_calendar_days)
        port_df = df.loc[
            (df[datevar] >= begin_date) &
            (df[datevar] <= extended_date) &
            (df[id_var].isin(port_assignments[id_var]))
            ]

        # Override with beginning port assignments
        port_df.drop(portvar, axis=1, inplace=True)
        port_df = port_df.merge(port_assignments, how='left', on=id_var)
        # Override with portfolio formation date
        port_df[portfolio_datevar] = port_date

        output_df = output_df.append(port_df)

    return output_df

def _portfolio_begin_date(port_date: pd.Timestamp, needed_prior_calendar_days: Optional[int] = None) -> pd.Timestamp:
    if needed_prior_calendar_days is None:
        return port_date

    return port_date - np.timedelta64(needed_prior_calendar_days, 'D')


import pandas as pd

def diff_df_from_panel_df(df: pd.DataFrame, entity_var: str, time_var: str, treated_var: str,
                          treated_time_var: str = 'After') -> pd.DataFrame:
    df = df.sort_values([entity_var, time_var])
    first_df = df.groupby(entity_var, as_index=False).first()
    first_df[treated_time_var] = 0
    last_df = df.groupby(entity_var, as_index=False).last()
    last_df[treated_time_var] = 1
    diff_df = first_df.append(last_df)

    return diff_df